#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/ctrManager.c,v 2.11 1996/09/05 15:05:45 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: ctrManager.c,v $
 * Revision 2.11  1996/09/05 15:05:45  richter
 * new 'tag' field (for the display list interface)
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/25  05:54:54  richter
 * handle vertac
 *
 * Revision 2.9.1.1  1996/07/03  10:06:43  richter
 * refined view scaling
 * configure output
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:05  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "ctrProto.h"
#include "namProto.h"
#include "pltProto.h"
#include "segProto.h"
#include "strProto.h"
#include "setProto.h"
#include "styProto.h"

ContourP
ctrAlloc(char *name)
{
    ContourP cPtr;

    cPtr = (ContourP) ckalloc( sizeof(ContourT) );
    if (! cPtr) {
	Tcl_AppendResult(interp, "memory allocation error", NIL);
	return NULL;
    }

    ctrClean(cPtr, FALSE);
    cPtr->name = strDupl(name);
    if (namAlloc("CTR", name, (void *)cPtr) != TCL_OK) {
        ctrFree(&cPtr);
        cPtr = NULL;
    }
    
    return(cPtr);
}

void
ctrClean(ContourP cPtr,
	 int old)
{
    if (!cPtr) return;

    styClean(&cPtr->style[0], old);
    styClean(&cPtr->style[1], old);

    if ( old ) {
	if (cPtr->name)   ckfree(cPtr->name);
	if (cPtr->level)  lvlFree(&cPtr->level);
	if (cPtr->view)   ckfree(cPtr->view);
	if (cPtr->object) ckfree(cPtr->object);
    }

    cPtr->name =   NULL;
    cPtr->flags =  0;
    cPtr->nl =     0;
    cPtr->vertac = 1.;
    cPtr->level =  NULL;
    cPtr->view =   NULL;
    cPtr->object = NULL;
    return;
}

void
ctrFree(ContourP *ctr)
{
    ContourP cPtr;

    cPtr = *ctr;
    if (!cPtr) return;

    namFree("CTR", cPtr->name);

    styClean(&cPtr->style[0], TRUE);
    styClean(&cPtr->style[1], TRUE);
    if (cPtr->level)  lvlFree(&cPtr->level);
    if (cPtr->name)   ckfree(cPtr->name);
    if (cPtr->view)    ckfree(cPtr->view);
    if (cPtr->object) ckfree(cPtr->object);
    ckfree( (char *)cPtr );
    *ctr = (ContourP) NULL;
    return;
}

int
ctrListByName(char *match)
{
    return namListByName("CTR", match);
}


ContourP
ctrFindByName(char *name)
{
    return (ContourP) namFindByName("CTR", name);
}


/*
 *----------------------------------------------------------------------
 * Level (lvl)
 *----------------------------------------------------------------------
 */
LevelP
lvlAlloc( void )
{
    LevelP lvl;

    lvl = (LevelP) ckalloc( sizeof(LevelT) );
    lvlClean(lvl, FALSE);
    return(lvl);
}

void
lvlClean(LevelP lvl,
	 int old)
{
    if (!lvl) return;

    if (old) {
	if (lvl->segments) segFree( &lvl->segments );
    } else {
	lvl->next = NULL;
    }
    lvl->ns       = 0;
    lvl->flags    = 0;
    lvl->tag      = 0;
    lvl->value    = 0.;
    lvl->segments = NULL;
    lvl->next     = NULL;
}

void
lvlFree( LevelP *lvl )
{
    LevelP l;

    l = *lvl;
    if (!l) return;
    if (l->segments) segFree( &l->segments );
    if (l->next) lvlFree( &l->next );
    ckfree( (char *)l );
    *lvl = (LevelP) NULL;
    return;
}

void
lvlClearPlotFlag( LevelP lvl )
{
    while (lvl) {
	lvl->flags &= ~ M_LVL_PLOT_DONE;
	lvl = lvl->next;
    }
    return;
}

void
lvlClearCalcFlag( LevelP lvl )
{
    while (lvl) {
	lvl->flags &= ~ M_LVL_CALC_DONE;
	lvl = lvl->next;
    }
    return;
}
