#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/dlManager.c,v 2.13 1996/11/11 17:45:20 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: dlManager.c,v $
 * Revision 2.13  1996/11/11 17:45:20  richter
 * casting on Tcl_GetHashValue
 *
 * Revision 2.12  1996/09/05  15:07:24  richter
 * completed display list interface
 *
 * Revision 2.11  1996/07/28  11:50:20  richter
 * handle three different drawing planes (background, main, foreground) by
 * inserting new display list elements at different parts of the list.
 *
 * Revision 2.10  1996/07/25  09:19:42  richter
 * *** empty log message ***
 *
 * Revision 1.1  1996/07/25  09:18:33  richter
 * Initial revision
 *
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"            /* type definitions and array sizes     */
#include "dlProto.h"
#include "strProto.h"
#include "styProto.h"

static DLobjP dlGet(int tag);
static void dlClean(DLobjP dlPtr, DLobjE type);

/*
 * variables with internal linkage
 */
static Tcl_HashTable dlHashTable;
static int dlHashTable_unused = 1;

int
dlAlloc(ViewP view,
        DLobjE type,
        DLplaneE plane)
{
    Tcl_HashEntry *hPtr = NULL;
    DLobjP dlPtr;
    static int last_tag = 0;
    int new;

    dlPtr = (DLobjP) ckalloc( sizeof(DLobjT) );
    if (!dlPtr) {
        return 0;
    }

    dlClean(dlPtr, type);

    if (view->dl_current == NULL) {
        view->dl_head =    dlPtr;
        view->dl_tail =    dlPtr;
        view->dl_current = dlPtr;
    } else {
        switch (plane) {
        case DL_BG_PLANE: {
            /*
             * background - prepend before dl_head
             */
            view->dl_head->next = dlPtr;
            dlPtr->next = view->dl_head;
            view->dl_head = dlPtr;
        }; break;
        case DL_FG_PLANE: {
            /*
             * foreground - append after dl_tail
             */
            view->dl_tail->next = dlPtr;
            dlPtr->prev = view->dl_tail;
            view->dl_tail = dlPtr;
        }; break;
        case DL_MAIN_PLANE: {
            DLobjP n;
            /*
             * main graphics - insert after dl_current
             */
            if ((n=view->dl_current->next)) {
                n->prev = dlPtr;
                dlPtr->next = n;
            }
            view->dl_current->next = dlPtr;
            dlPtr->prev = view->dl_current;
            if (view->dl_current == view->dl_tail) {
                view->dl_tail = dlPtr;
            }
            view->dl_current = dlPtr;
        }; break;
        }
    }

    dlPtr->tag = ++last_tag;
    
    if (dlHashTable_unused) {
	Tcl_InitHashTable(&dlHashTable, TCL_ONE_WORD_KEYS);
        dlHashTable_unused = 0;
    }
    hPtr = Tcl_CreateHashEntry(&dlHashTable, (char *) dlPtr->tag,
                               &new);
    if (! new) {
    	Tcl_AppendResult(interp, "(internal error): tag already exists", NIL);
	return TCL_ERROR;
    }

    Tcl_SetHashValue(hPtr, (ClientData) dlPtr);
    return dlPtr->tag;
}

static void
dlClean(DLobjP dlPtr,
        DLobjE type)
{
    dlPtr->type = type;
    dlPtr->tag  = 0;
    dlPtr->done = FALSE;
    dlPtr->object.nl = NULL;
    dlPtr->next = NULL;
    dlPtr->prev = NULL;
}

void
dlMkCylinder(int tag,
              c3dT *crd1,
              c3dT *crd2,
              double radius,
              StyleP style)
{
    DLobjP dlPtr;
    DLCylinder *o;
    c3dT d;
    
    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Cylinder = o = (DLCylinder *) ckalloc(sizeof(DLCylinder));
    o->radius = radius;
    o->origin.x = crd1->x;
    o->origin.y = crd1->y;
    o->origin.z = crd1->z;
    d.x = crd2->x - crd1->x;
    d.y = crd2->y - crd1->y;
    d.z = crd2->z - crd1->z;
    o->height = sqrt(d.x*d.x + d.y*d.y +d.z*d.z);
    o->axis.x = -d.y;
    o->axis.y =  d.x;
    o->axis.z = 0.;
    o->angle = 57.29578 * atan2(sqrt(d.x*d.x+d.y*d.y), d.z);
    return;
}

void
dlRmCylinder(DLCylinder *o)
{
    ckfree((char *) o);
    return;
}

void
dlMkHeight(int tag,
           char *map,
           int plane,
           double vertac,
           double hcut,
           double lcut,
           GridP gPtr,
           StyleP style)
{
    DLobjP dlPtr;
    DLHeight *h;
    int i, k, me, sign;
    double d1x, d1y, d1z;
    double d2x, d2y, d2z, z;
    c3dT *facet;

    sign = FALSE;
    
    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Height = h = (DLHeight *) ckalloc(sizeof(DLHeight));
    h->nx = gPtr->nx;
    h->ny = gPtr->ny;
    h->xinc = gPtr->xinc;
    h->yinc = gPtr->yinc;
    h->vertac = vertac;
    h->hcut = hcut;
    h->lcut = lcut;
    h->listid = 0;
    h->vertex = (c3dT *) ckalloc(h->nx*h->ny*sizeof(c3dT));
    if (sign) {
        h->color = (GtColor *) ckalloc(h->nx*h->ny*sizeof(GtColor));
    } else {
        h->color = NULL;
    }
    for (k=0; k < h->ny; ++k) {
        for (i=0; i < h->nx; ++i) {
            me = i + k*h->nx;
            h->vertex[me].x = i*h->xinc - gPtr->xran/2.;
            h->vertex[me].y = k*h->yinc - gPtr->yran/2.;
            z = h->vertex[me].z = h->vertac*gPtr->va[me + plane*h->ny*h->nx];
            if (sign) {
                if (z < -0.001) {
                    h->color[me].rgb[0] = 0.88;
                    h->color[me].rgb[1] = 0.55;
                    h->color[me].rgb[2] = 0.55;
                } else if (z > 0.001) {
                    h->color[me].rgb[0] = 0.55;
                    h->color[me].rgb[1] = 0.55;
                    h->color[me].rgb[2] = 1.00;
                }
            }
        }
    }

    facet = (c3dT *) ckalloc(h->nx*h->ny*sizeof(c3dT));
    for (k=0; k < (h->ny - 1); k++) {
        for (i=0; i < (h->nx - 1); i++) {
            me = i + k*h->nx;
            d1x = h->vertex[me+1].x     - h->vertex[me].x;
            d1y = h->vertex[me+1].y     - h->vertex[me].y;
            d1z = h->vertex[me+1].z     - h->vertex[me].z;
            d2x = h->vertex[me+h->nx].x - h->vertex[me].x;
            d2y = h->vertex[me+h->nx].y - h->vertex[me].y;
            d2z = h->vertex[me+h->nx].z - h->vertex[me].z;
            facet[i+k*h->nx].x = d1y*d2z - d1z*d2y;
            facet[i+k*h->nx].y = d1z*d2x - d1x*d2z;
            facet[i+k*h->nx].z = d1x*d2y - d1y*d2x;
        }
    }
    
    h->normal = (c3dT *) ckalloc(h->nx*h->ny*sizeof(c3dT));
    for (k=0; k < h->ny; k++) {
        int lrow, urow;
        if (k <= 0) {
            lrow = 0;
            urow = 0;
        } else if (k >= h->ny-1) {
            lrow = (h->ny - 2) * h->nx;
            urow = (h->ny - 2) * h->nx;
        } else {
            lrow = (k-1)*h->nx;
            urow = k * h->nx;
        }
        for (i=0; i < h->nx; i++) {
            int lcol, rcol;
            int me, ll, lr, ul, ur;

            if (i <= 0) {
                lcol = 0;
                rcol = 0;
            } else if (i >= h->nx-1) {
                lcol = h->nx - 2;
                rcol = h->nx - 2;
            } else {
                lcol = i - 1;
                rcol = i;
            }
            ll = lrow + lcol;
            lr = lrow + rcol;
            ul = urow + lcol;
            ur = urow + rcol;
            me = i + k * h->nx;
            h->normal[me].x = facet[ll].x + facet[lr].x +
                facet[ul].x + facet[ur].x;
            h->normal[me].y = facet[ll].y + facet[lr].y +
                facet[ul].y + facet[ur].y;
            h->normal[me].z = facet[ll].z + facet[lr].z +
                facet[ul].z + facet[ur].z;
        }
    }
    ckfree((char *)facet);

}

void
dlRmHeight(DLHeight *h)
{
    if (h->vertex) ckfree((char *) h->vertex);
    if (h->color)  ckfree((char *) h->color);
    if (h->normal) ckfree((char *) h->normal);
    ckfree((char *) h);
    return;
}

void
dlMkMarker(int tag,
           c3dT *crd,
           int marker,
           StyleP style)
{
    DLobjP dlPtr;
    DLMarker *o;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Marker = o = (DLMarker *) ckalloc(sizeof(DLMarker));
    o->marker = marker;
    o->vertex.x = crd->x;
    o->vertex.y = crd->y;
    o->vertex.z = crd->z;
    return;
}

void
dlRmMarker(DLMarker *o)
{
    ckfree((char *) o);
    return;
}

void
dlMkSegment(int tag,
            SegmentP segment,
            StyleP style)
{
    DLobjP dlPtr;
    DLSegment *o;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Segment = o = (DLSegment *) ckalloc(sizeof(DLSegment));
    o->v = segment;
    return;
}

void
dlRmSegment(DLSegment *o)
{
    ckfree((char *) o);
    return;
}

void
dlMkSphere(int tag,
           c3dT *crd,
           double radius,
           StyleP style)
{
    DLobjP dlPtr;
    DLSphere *o;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Sphere = o = (DLSphere *) ckalloc(sizeof(DLSphere));
    o->radius = radius;
    o->vertex.x = crd->x;
    o->vertex.y = crd->y;
    o->vertex.z = crd->z;
    return;
}

void
dlRmSphere(DLSphere *o)
{
    ckfree((char *) o);
    return;
}

void
dlMkText(int tag,
         c3dT *crd,
         char *string,
         StyleP style)
{
    DLobjP dlPtr;
    DLText *o;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Text = o = (DLText *) ckalloc(sizeof(DLText));
    o->string = strDupl(string);
    o->vertex.x = crd->x;
    o->vertex.y = crd->y;
    o->vertex.z = crd->z;
    return;
}

void
dlRmText(DLText *o)
{
    if (o->string) ckfree((char *) o->string);
    ckfree((char *) o);
    return;
}

void
dlMkTriangle(int tag,
             TriangleP tri,
             StyleP style)
{
    DLobjP dlPtr;
    DLTriangle *o;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;
    memcpy(&dlPtr->style, style, sizeof(StyleT));
    dlPtr->object.Triangle = o = (DLTriangle *) ckalloc(sizeof(DLTriangle));
    o->head = tri;
    return;
}

void
dlRmTriangle(DLTriangle *o)
{
    ckfree((char *) o);
    return;
}

static DLobjP
dlGet(int tag)
{
    Tcl_HashEntry *hPtr = NULL;
    DLobjP dlPtr;

    if (tag == 0) {
        return (DLobjP) NULL;
    }
    
    if (dlHashTable_unused) {
        printf("empty table\n");
        return (DLobjP) NULL;
    }
    hPtr = Tcl_FindHashEntry(&dlHashTable, (char *) tag);
    if (!hPtr) {
        printf("tag %d not found\n", tag);
        return (DLobjP) NULL;
    }
    
    dlPtr = (DLobjP) Tcl_GetHashValue(hPtr);
    return dlPtr;
}

void
dlRemove(ViewP view,
         int tag)
{
    DLobjP dlPtr;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;

    pltRemove(view, tag);
    
    /*
     * remove from display list
     */
    if (dlPtr->prev) {
        dlPtr->prev->next = dlPtr->next;
    }
    if (dlPtr->next) {
        dlPtr->next->prev = dlPtr->prev;
    }
    /*
     * correct viewing plane pointer
     */
    if (view->dl_head == dlPtr) {
        view->dl_head = dlPtr->next;
    }
    if (view->dl_tail == dlPtr) {
        view->dl_tail = dlPtr->prev;
    }
    if (view->dl_current == dlPtr) {
        if (dlPtr->prev) {
            view->dl_current = dlPtr->prev;
        } else {
            view->dl_current = dlPtr->next;
        }
    }
    /*
     * clean
     */
    if (dlPtr->object.nl) {
        switch (dlPtr->type) {
        case DL_SEGMENT:
            dlRmSegment(dlPtr->object.Segment);
            break;
        case DL_TRIANGLE:
            dlRmTriangle(dlPtr->object.Triangle);
            break;
            /*        DL_RECTANGLE: */
        case DL_HEIGHTFIELD:
            dlRmHeight(dlPtr->object.Height);
            break;
        case DL_SPHERE:
            dlRmSphere(dlPtr->object.Sphere);
            break;
        case DL_CYLINDER:
            dlRmCylinder(dlPtr->object.Cylinder);
            break;
        case DL_TEXT:
            dlRmText(dlPtr->object.Text);
            break;
        case DL_MARKER:
            dlRmMarker(dlPtr->object.Marker);
            break;
        }
    }
    ckfree((char *) dlPtr);
    return;
}

void
dlRemoveData(ViewP view,
             int tag)
{
    DLobjP dlPtr;

    dlPtr = dlGet(tag);
    if (!dlPtr) return;

    /*
     * cleanup data for some dl types
     */
    if (dlPtr->object.nl) {
        if (dlPtr->type == DL_SEGMENT) {
            ckfree((char *) dlPtr->object.Segment->v);
            dlPtr->object.nl = NULL;
        }
    }
    return;
}
