#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/drvGT.c,v 2.12 1999/08/20 14:01:51 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: drvGT.c,v $
 * Revision 2.12  1999/08/20 14:01:51  richter
 * Some compiler complained about 'i = ++i % npat'.  Something about
 * modifying i twice without intervening sequence point.  Fixed by
 * splitting it into two statements (ugh!).
 *
 * Revision 2.11  1996/09/05 15:09:15  richter
 * moved a huge part into the generic display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:02:07  richter
 * optionally map sign on heightfields
 *
 * Revision 2.9.1.2  1996/07/03  10:10:41  richter
 * fixed pattern to stipple conversion
 * fixed taglists (delimiter is now above ASCII 127)
 *
 * Revision 2.9.1.1  1996/05/19  16:09:02  richter
 * new routines for bonds
 * load some bitmaps
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:06  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: drvGT
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#ifdef WIN32                    /* Cljf for Windows compilation         */
#include <windows.h>
#endif
#include "conf.h"		/* global configuration			*/
#include <tk.h>			/* Tk interface				*/
#include <GL/gl.h>
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "strProto.h"
#include "drvProto.h"
#include "bitmaps/onoff.xbm"
#include "bitmaps/default.xbm"
#include "bitmaps/undo.xbm"
#include "bitmaps/choose.xbm"
#include "bitmaps/select.xbm"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */
extern Tcl_Interp *interp;	/* the interpreter 			*/
extern char buf[1000];

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static Tk_Window mainWindow;    /* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */
static char tag_delimiter = 0xa6;

void
gtInit(char *init)
{
    char *msg;
    int code;
    static char pltinit[] =
        "if [file exists $xd_library/xdGT.tcl] {\n\
source $xd_library/xdGT.tcl\n\
} else {\n\
set msg \"can't find $xd_library/xdGT.tcl; perhaps you \"\n\
append msg \"need to\\ninstall XD or set your XD_TCLDIR \"\n\
append msg \"environment variable?\"\n\
error $msg\n\
}";

    /*
     * Initialize the Tk application.
     */
#if XD_TK_VERSION <=40
    mainWindow = Tk_CreateMainWindow(interp, NULL, "xdgraph", "XD");
    if (mainWindow == NULL) {
	fprintf(stderr,"Can't open main window\n");
    }
#endif
    if ( (code = Tk_Init(interp)) == TCL_ERROR) {
        goto error;
    }
#if XD_TK_VERSION >= 41
    mainWindow = Tk_MainWindow(interp);
#endif
    
    if (TkGLx_Init(interp, mainWindow) == TCL_ERROR) {
        goto error;
    }

    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_onoff"), (char *) onoff_bits,
                        onoff_width, onoff_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_default"), (char *) default_bits,
                        default_width, default_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_undo"), (char *) undo_bits,
                        undo_width, undo_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_choose"), (char *) choose_bits,
                        choose_width, choose_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_select"), (char *) select_bits,
                        select_width, select_height) != TCL_OK) goto error;

    code = Tcl_Eval(interp, pltinit);
    if (code != TCL_OK) {
	goto error;
    }

    code = Tcl_Eval(interp, init);
    if (code != TCL_OK) {
	goto error;
    }

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we clean up and
     * return.
     */
    Tk_MainLoop();
    exit(0);

error:
    msg = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) {
        msg = interp->result;
    }
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(interp, "destroy .");
    exit(1);
}

int
gtClearView(DrvgtP ddp)
{
    return TCL_OK;
}

void
gtAllocDriverData(DrvgtP *ddp)
{
    int i, k;

    *ddp = (void *) ckalloc( sizeof(DrvgtT) );
    (*ddp)->xang = 0.;
    (*ddp)->yang = 0.;
    (*ddp)->xlast = 0;
    (*ddp)->ylast = 0;
    (*ddp)->pixw = 0;
    (*ddp)->pixh = 0;
    for (i=0; i<5; i++) {
        for (k=0; k<4; k++) {
            (*ddp)->light[i].position[k] = 0.;
            (*ddp)->light[i].ambient[k]  = 0.;
            (*ddp)->light[i].diffuse[k]  = 0.;
            (*ddp)->light[i].specular[k] = 0.;
        }
    }
    (*ddp)->lightnum = 2;
    (*ddp)->light[0].position[0] = 100.;
    (*ddp)->light[0].position[2] = 100.;
    (*ddp)->light[0].diffuse[0] = 0.5;
    (*ddp)->light[0].diffuse[1] = 0.5;
    (*ddp)->light[0].diffuse[2] = 0.4;
    (*ddp)->light[0].diffuse[3] = 1.0;
    (*ddp)->light[1].position[1] = 100.;
    (*ddp)->light[1].position[2] = 100.;
    (*ddp)->light[1].diffuse[0] = 0.4;
    (*ddp)->light[1].diffuse[1] = 0.5;
    (*ddp)->light[1].diffuse[2] = 0.5;
    (*ddp)->light[1].diffuse[3] = 1.0;
    for (i=0; i<15; i++) {
        (*ddp)->rotmat[i] = 0.;
    }
    (*ddp)->rotmat[0] = 1.;
    (*ddp)->rotmat[5] = 1.;
    (*ddp)->rotmat[10] = 1.;
    (*ddp)->rotmat[15] = 1.;
}

GtColor *
gtSetColorRGB(float rgb[3])
{
    GtColor *c;

    c = (GtColor *) ckalloc(sizeof(GtColor));
    c->rgb[0] = rgb[0];
    c->rgb[1] = rgb[1];
    c->rgb[2] = rgb[2];
    return c;
}

GtLinewidth *
gtSetLinewidth(double width)
{
    GtLinewidth *l;

    l = (GtLinewidth *) ckalloc(sizeof(GtLinewidth));
    l->width = width;
    return l;
}

GtLinestipple *
gtSetLinestipple(int plength,
                 double *pattern)
{
    GtLinestipple *l;
    int i, k, *bits, nbit, irep, *map, npat, itmp;
    int factor;
    double rf;
    unsigned short mask;

    mask = 0;
    factor = 1;
    map = NULL;
    bits = NULL;
    if (plength <= 0) {
        factor = 0;
    } else {
        npat = plength;
        bits = (int *) ckalloc(npat*sizeof(int));
        nbit = 0;
        for (i=0; i<plength; ++i) {
            bits[i] = (int) (pattern[i]/0.03);
            nbit += abs(bits[i]);
        }
        irep = 16/nbit;
        if (irep > 1) {
            npat *= irep;
            nbit *= irep;
            bits = (int *)ckrealloc((char *)bits, npat*sizeof(int));
            for (i=plength; i<npat; ++i) {
                bits[i] = bits[i%plength];
            }
        } else if (nbit > 16) {
            rf = ceil((double)nbit/16.);
            factor = (int) (rf + 0.5);
            nbit = 0;
            for (i=0; i<npat; ++i) {
                itmp = bits[i] > 0 ? 1 : -1;
                bits[i] = (int) ((double)bits[i]/rf + 0.5 * (double)itmp);
                if (bits[i] == 0) {
                    bits[i] += itmp;
                }
                nbit += abs(bits[i]);
            }
        }
        map = (int *) ckalloc(npat*sizeof(int));
        for (i=0; i<npat; ++i) {
            map[i] = i;
        }
        for (i=0; i<npat-1; ++i) {
            for (k=i+1; k<npat; ++k) {
                if (abs(bits[map[k]]) > abs(bits[map[i]])) {
                    itmp = map[k];
                    map[k] = map[i];
                    map[i] = itmp;
                }
            }
        }
        if (nbit < 16) {
            itmp = 16 - nbit;
            for (i=0; itmp > 0; --itmp) {
                k = map[i];
                bits[k] < 0 ? --bits[k] : ++bits[k];
                ++i;
                i %= npat;
            }
        } else if (nbit > 16) {
            itmp = nbit - 16;
            for (i=0; itmp > 0; --itmp) {
                k = map[i];
                bits[k] < 0 ? ++bits[k] : --bits[k];
                ++i;
                i %= npat;
            }
        }
        mask = 0;
        for (k=npat-1; k>=0; --k) {
            if (bits[k] > 0) {
                for (i=bits[k]; i>0; --i) {
                    mask |= 1;
                    mask <<= 1;
                }
            } else {
                mask <<= -bits[k];
            }
        }
    }
    if (bits) ckfree((char *)bits);
    if (map) ckfree((char *)map);
    l = (GtLinestipple *) ckalloc(sizeof(GtLinestipple));
    l->mask = mask;
    l->factor = factor;
    return l;
}

void
gtSetRotation(ViewP view,
              int xnow,
              int ynow)
{
    DrvgtP ddp;
    ddp = (DrvgtP) view->drdat;
    ddp->xang = (ynow - ddp->ylast);
    ddp->yang = (xnow - ddp->xlast);
    
    ddp->xlast = xnow;
    ddp->ylast = ynow;
}

void
gtSetRotx(ViewP view,
          int xang)
{
    DrvgtP ddp;
    ddp = (DrvgtP) view->drdat;
    ddp->xang += xang;
}

void
gtSetRoty(ViewP view,
          int yang)
{
    DrvgtP ddp;
    ddp = (DrvgtP) view->drdat;
    ddp->yang += yang;
    
}

void
gtSetOrigin(ViewP view,
              int xnow,
              int ynow)
{
    DrvgtP ddp;
    ddp = (DrvgtP) view->drdat;
    ddp->xlast = xnow;
    ddp->ylast = ynow;
}

void
gtSetWindowSize(ViewP view,
                int pixw,
                int pixh)
{
    DrvgtP ddp;
    ddp = (DrvgtP) view->drdat;
    ddp->pixw = pixw;
    ddp->pixh = pixh;
}
