#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/drvTK.c,v 2.17 1997/03/29 18:05:20 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: drvTK.c,v $
 * Revision 2.17  1997/03/29 18:05:20  richter
 * added static declarations where appropriate
 * handle the error code returned by tkFlush
 *
 * Revision 2.16  1996/11/29  14:31:32  richter
 * added some XD_TK_VERSION #ifs
 *
 * Revision 2.15  1996/11/03  16:31:37  richter
 * preset linestyle in tkRedraw, changing views got confused about this
 *
 * Revision 2.14  1996/09/11  11:21:24  richter
 * removed some debug output
 *
 * Revision 2.13  1996/09/06  17:50:55  richter
 * tkCheckAbort now uses Tcl_GetBolean
 * added tkSetAbort, used in tkRemove
 * in tkRedraw, use tkCheckAbort to avoid using objects removed during 'update'
 * propagate tkCheckAbort from tkFlush upwards using TCL_BREAK as the signal
 *
 * Revision 2.12  1996/09/06  13:25:53  richter
 * add tag 'label' for text items, so the can be moved interactively
 * use braces around text string instead of quote marks
 *
 * Revision 2.11  1996/09/05  15:10:48  richter
 * use the new display lists
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.1  1996/05/20  11:17:01  richter
 * new move/draw routines
 * load bitmaps
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:06  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: drvTk
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration			*/
#include <tk.h>			/* Tk interface				*/
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "strProto.h"
#include "drvProto.h"
#include "pltProto.h"
#include "styProto.h"
#include "bitmaps/onoff.xbm"
#include "bitmaps/default.xbm"
#include "bitmaps/undo.xbm"
#include "bitmaps/choose.xbm"
#include "bitmaps/select.xbm"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static char *tohex(int r, int g, int b);
static void tkPollEvent(void);
static int tkCheckAbort(void);
static void tkClearAbort(void);
static void tkSetAbort(void);
static int tkSetFillByName(DrvtkP ddp, char *color);
static int tkSetFillByRGB(DrvtkP ddp, float rgb[3]);
static int tkSetLinewidth(DrvtkP ddp, double lw);
static int tkSetFont(DrvtkP ddp, double size, char *name);
static int tkSetAnchor(DrvtkP ddp, char *anchor);
static int tkSetTag(DrvtkP ddp, int tag);
static int tkDrawMarker(ViewP view, DrvtkP ddp, DLMarker *m);


/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */
extern Tcl_Interp *interp;	/* the interpreter 			*/

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static Tk_Window mainWindow;    /* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */

static Tcl_DString plotcmd;

void
tkInit(char *init)
{
    char *msg;
    int code;
    static char pltinit[] =
        "if [file exists $xd_library/xdTK.tcl] {\n\
            source $xd_library/xdTK.tcl\n\
        } else {\n\
            set msg \"can't find $xd_library/xdTK.tcl; perhaps you \"\n\
            append msg \"need to\\ninstall XD or set your XD_TCLDIR \"\n\
            append msg \"environment variable?\"\n\
            error $msg\n\
        }";

    /*
     * Initialize the Tk application.
     */
#if XD_TK_VERSION <=40
    mainWindow = Tk_CreateMainWindow(interp, NULL, "xdgraph", "XD");
    if (mainWindow == NULL) {
	fprintf(stderr,"Can't open main window\n");
    }
#endif
    if ( (code = Tk_Init(interp)) == TCL_ERROR) {
        goto error;
    }
#if XD_TK_VERSION >=41
    mainWindow = Tk_MainWindow(interp);
#endif
    
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_onoff"), (char *) onoff_bits,
                        onoff_width, onoff_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_default"), (char *) default_bits,
                        default_width, default_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_undo"), (char *) undo_bits,
                        undo_width, undo_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_choose"), (char *) choose_bits,
                        choose_width, choose_height) != TCL_OK) goto error;
    if (Tk_DefineBitmap(interp, Tk_GetUid("xd_select"), (char *) select_bits,
                        select_width, select_height) != TCL_OK) goto error;
    
    code = Tcl_Eval(interp, pltinit);
    if (code != TCL_OK) {
	goto error;
    }

    Tcl_DStringInit(&plotcmd);

    code = Tcl_Eval(interp, init);
    if (code != TCL_OK) {
	goto error;
    }

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we clean up and
     * return.
     */
    Tk_MainLoop();
    Tcl_DStringFree(&plotcmd);
    exit(0);

error:
    msg = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) {
        msg = interp->result;
    }
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(interp, "destroy .");
    exit(1);
}

static void
tkPollEvent(void)
{
#if XD_TK_VERSION >= 41
    Tcl_DoOneEvent(TK_X_EVENTS|TK_DONT_WAIT);
#else
    Tk_DoOneEvent(TK_X_EVENTS|TK_DONT_WAIT);
#endif
}

static int
tkCheckAbort(void)
{
    char *res;
    int abort = 0;
    
    res = Tcl_GetVar(interp, "abort", TCL_GLOBAL_ONLY);
    if (res) {
        if (Tcl_GetBoolean(interp, res, &abort) != TCL_OK) {
            fprintf(stderr, "GetBoolean failed for '%s'\n", res);
        }
        return abort;
    } else {
	return 0;
    }
}

static void
tkClearAbort(void)
{
    Tcl_SetVar(interp, "abort", "0", TCL_GLOBAL_ONLY);
}

static void
tkSetAbort(void)
{
    Tcl_SetVar(interp, "abort", "1", TCL_GLOBAL_ONLY);
}

void
tkAllocDriverData(DrvtkP *ddp)
{
    *ddp = (void *) ckalloc( sizeof(DrvtkT) );
    (*ddp)->yoffset = 0.;
    (*ddp)->TheFill = NULL;
    (*ddp)->TheFont = NULL;
    (*ddp)->TheWidth = NULL;
    (*ddp)->TheAnchor = NULL;
    (*ddp)->TheTag = NULL;
}

void
tkRedraw(ViewP view,
         DLobjP o)
{
    int mask;
    DrvtkP ddp;
    StyleT style = {
        {NULL, UNKNOWNMODEL, {0., 0., 0.}},
        {0., 0, NULL},
        {0., NULL, NULL}
    };
    
    ddp = (DrvtkP) view->drdat;
    pltSetLinestyle(&view->style.line);
    tkClearAbort();
    
    for (; o; o=o->next) {
        if (o->done) continue;
        tkSetTag(ddp, o->tag);
        styClean(&style, TRUE);
        mask = 0;
        o->done = TRUE; 
        switch (o->type) {
        case DL_HEIGHTFIELD:
        case DL_TRIANGLE: 
        case DL_SPHERE: 
        case DL_CYLINDER: 
            break;
        case DL_SEGMENT:
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL|STY_LINE_ALL);
            if (mask) tkApplyStyle(ddp, mask, &style);
            tkDrawSegment(view, ddp, o->object.Segment);
            break;
        case DL_MARKER:
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL);
            if (mask) tkApplyStyle(ddp, mask, &style);
            tkDrawMarker(view, ddp, o->object.Marker);
            break;
        case DL_TEXT: 
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL|STY_FONT_ALL);
            if (mask) tkApplyStyle(ddp, mask, &style);
            tkDrawText(view, ddp, o->object.Text);
            break;
        }
        if (tkCheckAbort()) {
            printf("abort\n");
            break;
        }
   }
    tkClearAbort();
}


void
tkSetYoff(DrvtkP ddp,
	  double h)
{
    ddp->yoffset = h;
    return;
}

static int
tkSetFillByName(DrvtkP ddp,
		char *color)
{
    int code;
    if ((code = tkFlush(ddp, NULL, 0)) != TCL_OK) {
        return code;
    }
    if (ddp->TheFill) ckfree(ddp->TheFill);
    if (color) {
	ddp->TheFill = strDupl(color);
    } else {
	ddp->TheFill = NIL;
    }
    return TCL_OK;
}

static int
tkSetFillByRGB(DrvtkP ddp,
	       float rgb[3])
{
    int i;
    int code;
    
    if ((code = tkFlush(ddp, NULL, 0)) != TCL_OK) {
        return code;
    }
    if (ddp->TheFill) ckfree(ddp->TheFill);
    for (i=0; i<3; ++i) {
	if (rgb[i] > 1.) {
	    rgb[i] = 1.;
	} else if (rgb[i] < 0.) {
	    rgb[i] = 0.;
	}
    }
    ddp->TheFill = tohex((int)(255.*rgb[0]), (int)(255.*rgb[1]),
			 (int)(255.*rgb[2]));
    return TCL_OK;
}

static int
tkSetLinewidth(DrvtkP ddp,
	       double lw)
{
    int code;
    
    if ((code = tkFlush(ddp, NULL, 0)) != TCL_OK) {
        return code;
    }
    if (ddp->TheWidth) ckfree(ddp->TheWidth);
    if (lw > 0.) {
	sprintf(buf, "%.2f", lw);
	ddp->TheWidth = strDupl(buf);
    } else {
	ddp->TheWidth = NULL;
    }
    return TCL_OK;
}

static int
tkSetFont(DrvtkP ddp,
	  double size,
	  char *name)
{
    static char *pstable[] = {
	"Helvetica",
	"Helvetica-Bold"
    };
    static char *tktable[] = {
	"-Adobe-Helvetica-Medium-R-Normal--*-%d-*",
	"-Adobe-Helvetica-Bold-R-Normal--*-%d-*"
    };

    static const int defid = 1;
    static const int defpt = 100;

    int id, pt;
    int code;
    
    if ((code = tkFlush(ddp, NULL, 0)) != TCL_OK) {
        return code;
    }

    if (size > 0.) {
	pt = (int) (size*10.);
    } else {
	pt = defpt;
    }

    if (name != NULL) {
	id = 0;
	while (id<defid && strcmp(name, pstable[id]) == 0) ++id;
    } else {
	id = defid;
    }

    if (ddp->TheFont) ckfree(ddp->TheFont);
    sprintf(buf, tktable[id], pt);
    ddp->TheFont = strDupl(buf);
    return TCL_OK;
}

static int
tkSetAnchor(DrvtkP ddp,
	    char *anchor)
{
    int code;
    
    if ((code = tkFlush(ddp, NULL, 0)) != TCL_OK) {
        return code;
    }
    if (ddp->TheAnchor) ckfree(ddp->TheAnchor);
    if (anchor) {
	ddp->TheAnchor = strDupl(anchor);
    } else {
	ddp->TheAnchor = NULL;
    }
    return TCL_OK;
}

static int
tkSetTag(DrvtkP ddp,
         int tag)
{
    char tagstring[32];
    int code;
    
    if ((code = tkFlush(ddp, NULL, 0)) != TCL_OK) {
        return code;
    }
    if (tag == 0) return TCL_OK;
    sprintf(tagstring, "dl%6.6d", tag);
    ddp->TheTag = strDuplOver(ddp->TheTag, tagstring);
    return TCL_OK;
}

static int
tkDrawMarker(ViewP view,
             DrvtkP ddp,
             DLMarker *m)
{
    MarkDataT *mark;
    void *d;
    c3dT a;
    
    pltTransform(view, &m->vertex, &a);

    mark = pltGetMarker(m->marker);
    if (mark->mode == -5) {
        double xll, yll;
        xll = a.x + mark->x;
        yll = a.y + mark->y;
        mark++;
        tkOutlineRectangle(ddp, xll, yll, a.x + mark->x, a.y + mark->y);
    } else if (mark->mode == -1) {
        d = (void *)ddp;
        tkMove(d, a.x+mark->x, a.y+mark->y);
        for (mark++; mark->mode >= 0; ++mark) {
            if (mark->mode == 0) {
                tkMove(d, a.x+mark->x, a.y+mark->y);
            } else {
                tkDraw(d, a.x+mark->x, a.y+mark->y);
            }
        }
    }
    return TCL_OK;
}

int
tkMark(DrvtkP ddp,
       double x,
       double y,
       MarkDataT *mark)
{
    void *d;

    d = (void *)ddp;
    tkMove(d, x+mark->x, y+mark->y);
    for (mark++; mark->mode >= 0; ++mark) {
        if (mark->mode == 0) {
            tkMove(d, x+mark->x, y+mark->y);
        } else {
            tkDraw(d, x+mark->x, y+mark->y);
        }
    }
    return TCL_OK;
}

int
tkDrawSegment(ViewP view,
              DrvtkP ddp,
              DLSegment *p)
{
    int code;
    SegmentP s;

    if (!p) {
        fprintf(stderr, "nil seg.pointer\n");
        return TCL_ERROR;
    }
    for (s=p->v; s; s=s->next) {
        if ( s->type == INVALIDSEG || s->np <= 0 ) {
            continue;
        }
        switch (s->type) {
        case MOVESEG:
        case DRAWSEG:
        case LINESET: {
            code = pltSegment(view, s, tkMove, tkDraw);
            if (code == TCL_BREAK) {
                return TCL_OK;
            } else if (code != TCL_OK) {
                return code;
            }
        }; break;
        case POINTSET: {
            tkDrawPointSet(view, ddp, s);
        }; break;
        }
        if (tkCheckAbort()) return TCL_OK;
    }
    return TCL_OK;
}

void
tkDrawPointSet(ViewP view,
               DrvtkP ddp,
               SegmentP s)
{
    int n;
    c3dT a;
    
    for ( n=0; n < s->np; n++ ) {
        pltTransform(view, &s->p.point[n], &a);
        tkMark(ddp, a.x, a.y, pltGetMarker(3));
    }
    return;
}
    
int
tkDrawText(ViewP view,
           DrvtkP ddp,
           DLText *t)
{
    int code, len;
    c3dT a;
    
    pltTransform(view, &t->vertex, &a);

    if ((code=tkFlush(ddp, NIL, 0)) != TCL_OK) return code;

    Tcl_DStringAppend(&plotcmd, "$plotcmd create text", -1);

    len = sprintf(buf, " %.3fc %.3fc", a.x, ddp->yoffset-a.y) ;
    Tcl_DStringAppend(&plotcmd, buf, len );
    if (ddp->TheAnchor) {
	Tcl_DStringAppend(&plotcmd, " -anchor ", -1);
	Tcl_DStringAppend(&plotcmd, ddp->TheAnchor, -1);
    }
    if (ddp->TheFont) {
	Tcl_DStringAppend(&plotcmd, " -font ", -1);
	Tcl_DStringAppend(&plotcmd, ddp->TheFont, -1);
    }
    Tcl_DStringAppend(&plotcmd, " -text {", -1);
    Tcl_DStringAppend(&plotcmd, t->string, -1);
    Tcl_DStringAppend(&plotcmd, "}", 1);
    code = tkFlush(ddp, NIL, 1);
    
    if (code == TCL_OK && ddp->TheTag) {
        Tcl_DStringAppend(&plotcmd, "$plotcmd addtag label withtag {", -1);
        Tcl_DStringAppend(&plotcmd, ddp->TheTag , -1);
        Tcl_DStringAppend(&plotcmd, "}", 1);
        code = Tcl_GlobalEval(interp, Tcl_DStringValue(&plotcmd) );
        Tcl_DStringTrunc(&plotcmd, 0);
    }
    return code;
}

int
tkRectangle(DrvtkP ddp,
	    double xll,
	    double yll,
	    double xur,
	    double yur)
{
    int len, code;

    if ((code=tkFlush(ddp, NIL, 0)) != TCL_OK) return code;
    
    Tcl_DStringAppend(&plotcmd, "$plotcmd create rectangle", -1);

    len = sprintf(buf, " %.3fc %.3fc", xll, ddp->yoffset-yll) ;
    Tcl_DStringAppend(&plotcmd, buf, len );
    len = sprintf(buf, " %.3fc %.3fc", xur, ddp->yoffset-yur) ;
    Tcl_DStringAppend(&plotcmd, buf, len );
    Tcl_DStringAppend(&plotcmd, " -outline ", -1);
    Tcl_DStringAppend(&plotcmd, "\"\"", -1);

    return tkFlush(ddp, NIL, 0);
}

int
tkOutlineRectangle(DrvtkP ddp,
                   double xll,
                   double yll,
                   double xur,
                   double yur)
{
    int len, code;

    if ((code=tkFlush(ddp, NIL, 0)) != TCL_OK) return code;
    
    Tcl_DStringAppend(&plotcmd, "$plotcmd create rectangle", -1);

    len = sprintf(buf, " %.3fc %.3fc", xll, ddp->yoffset-yll) ;
    Tcl_DStringAppend(&plotcmd, buf, len );
    len = sprintf(buf, " %.3fc %.3fc", xur, ddp->yoffset-yur) ;
    Tcl_DStringAppend(&plotcmd, buf, len );

    return tkFlush(ddp, NIL, 0);
}

int
tkMove(void *drdat,
       double x,
       double y)
{
    DrvtkP ddp = (DrvtkP) drdat;
    sprintf(buf, "$plotcmd create line %.3fc %.3fc", x, ddp->yoffset-y );
    return tkFlush(ddp, buf, 0);
}

int
tkDraw(void *drdat,
       double x,
       double y)
{
    int len;
    DrvtkP ddp = (DrvtkP) drdat;

    len = sprintf(buf, " %.3fc %.3fc", x, ddp->yoffset-y);
    Tcl_DStringAppend( &plotcmd, buf, len );
    return TCL_OK;
}

int
tkFlush(DrvtkP ddp,
        char *init,
        int what)
{
    int code;
    static int length = 0;

    if ( Tcl_DStringLength( &plotcmd ) > length ) {
	if (ddp->TheFill) {
	    Tcl_DStringAppend(&plotcmd, " -fill ", -1);
	    if (strcasecmp(ddp->TheFill,"Blank")==0) {
		Tcl_DStringAppend(&plotcmd, "\"\"", -1);
	    } else {
		Tcl_DStringAppend(&plotcmd, ddp->TheFill, -1);
	    }
	}
	if (ddp->TheWidth && what != 1) {
	    Tcl_DStringAppend(&plotcmd, " -width ", -1);
	    Tcl_DStringAppend(&plotcmd, ddp->TheWidth, -1);
	}
	if (ddp->TheTag && strlen(ddp->TheTag)) {
	    Tcl_DStringAppend(&plotcmd, " -tags {", -1);
	    Tcl_DStringAppend(&plotcmd, ddp->TheTag, -1);
	    Tcl_DStringAppend(&plotcmd, "}", -1);
	}
        /*
	printf("%s\n", Tcl_DStringValue(&plotcmd) );
        /*	*/
	code = Tcl_GlobalEval(interp, Tcl_DStringValue(&plotcmd) );
	if (code == TCL_OK) {
            code = Tcl_Eval(interp, "update");
	}
    } else {
	code = TCL_OK;
    }

    length = 0;
    Tcl_DStringTrunc(&plotcmd, length);

    if (init) {
	Tcl_DStringAppend(&plotcmd, init, -1);
	length = Tcl_DStringLength(&plotcmd);
    }
    if (code == TCL_ERROR) {
	Tcl_AddErrorInfo(interp, "(tkFlush)\n");
    }

    if (tkCheckAbort()) return TCL_BREAK;
    return code;
}


static char *
tohex(int r,
      int g,
      int b)
{
    static char m[] = { "0123456789abcdef" };
    char *h;
    int vh, vl;

    h = ckalloc(8);
    h[0] = '#';
    vh = (r & 0xF0) >> 4;
    vl = (r & 0x0F);
    h[1] = m[vh];
    h[2] = m[vl];
    vh = (g & 0xF0) >> 4;
    vl = (g & 0x0F);
    h[3] = m[vh];
    h[4] = m[vl];
    vh = (b & 0xF0) >> 4;
    vl = (b & 0x0F);
    h[5] = m[vh];
    h[6] = m[vl];
    h[7] = '\000';

    return h;
}

int
tkApplyStyle(DrvtkP ddp,
             int mask,
             StyleP style)
{
    if (mask&STY_FG_ALL) {
        static char blank[] = "Blank";
        float rgb[3];
        int by_name = 0;
        char *name;
        ColorP cPtr = &style->fg;

        switch (cPtr->model) {
        case RGBMODEL:
            if (cPtr->name) {
                by_name = 1;
                name = cPtr->name;
            }
            rgb[0] = cPtr->val[0];
            rgb[1] = cPtr->val[1];
            rgb[2] = cPtr->val[2];
            break;
        case HSVMODEL:
            styCvthsv2rgb((float) cPtr->val[0]/360.,
                          cPtr->val[1], cPtr->val[2],
                          &rgb[0], &rgb[1], &rgb[2]);
            break;
        case YUVMODEL:
            styCvtyuv2rgb(cPtr->val[0], cPtr->val[1], cPtr->val[2],
                          &rgb[0], &rgb[1], &rgb[2]);
            break;
        case MAPMODEL: 
            {
                if (!styCvtindex2rgb(0, (int) cPtr->val[1],
                                     &rgb[0], &rgb[1], &rgb[2])) {
                    by_name = 1;
                    name = blank;
                }
            }
            break;
        }
        if (by_name) {
            tkSetFillByName(ddp, name);
        } else {
            tkSetFillByRGB(ddp, rgb);
        }
    }

    if (mask&STY_LINE_WIDTH) {
        tkSetLinewidth(ddp, style->line.width);
    }

    if (mask&STY_LINE_PATTERN) {
        pltSetLinestyle(&style->line);
    }

    if (mask&STY_FONT_NAME || mask&STY_FONT_SIZE) {
        if (style->font.name) {
            tkSetFont(ddp, style->font.size, style->font.name);
        } else {
            tkSetFont(ddp, -1., NULL);
        }
    }
    
    if (mask&STY_FONT_NAME) {
        tkSetAnchor(ddp, style->font.anchor);
    }
    return TCL_OK;
}

int
tkRemove(ViewP view,
         int tag)
{
    char tagstring[32];

    if (tag == 0) return TCL_OK;
    sprintf(tagstring, "dl%6.6d", tag);
    tkSetAbort();
    return Tcl_VarEval(interp, "ClearObject ", view->name, " ", tagstring,
                       NULL);
}
