#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/heiCommand.c,v 2.14 1997/03/29 18:05:46 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 *======================================================================
 *
 * Module: heiCommand
 *
 * Abstract: Defines the 'height' command and creates height object
 * commands.  Called from the tcl interpreter.
 *
 * $Log: heiCommand.c,v $
 * Revision 2.14  1997/03/29 18:05:46  richter
 * preset style
 *
 * Revision 2.13  1997/02/07  16:02:27  richter
 * *** empty log message ***
 *
 * Revision 2.12  1996/11/03  16:41:00  richter
 * styGetConfig expects a style group mask
 * added a missing Tcl_DStringEndSublist()
 *
 * Revision 2.11  1996/09/06  11:19:46  richter
 * many changes to use the new display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  05:50:16  richter
 * use Tcl_SetVar2 to set active(heightfield)
 * handle 'heightfield <height>' correctly
 * cmdSetView takes an array of possible view types
 * new options -cutoff and -map
 *
 * Revision 2.9.1.2  1996/07/03  09:55:44  richter
 * refined view scaling
 * configure output
 *
 * Revision 2.9.1.1  1996/05/19  15:54:57  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:04  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "cmdProto.h"
#include "heiProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "setProto.h"
#include "strProto.h"
#include "styProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureHeight(Tcl_Interp *interp, HeightP hPtr,
			    int argc,char **argv, ViewOptT *vOpt, int append);
static int showcfgHeight(Tcl_Interp *interp, HeightP hPtr);
static int infoHeight(Tcl_Interp *interp, HeightP hPtr,
		       int argc,char **argv);
static int plotHeight(Tcl_Interp *interp, ViewP view, HeightP hPtr);
static int clearHeight(Tcl_Interp *interp, ViewP view, HeightP hPtr);

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
HeightCmd(ClientData dummy,		/* Not used			*/
	   Tcl_Interp *interp,		/* Current interpreter		*/
	   int argc,			/* Number of arguments		*/
	   char **argv)			/* Argument strings		*/
{
    DatasetP dPtr;
    HeightP hPtr;
    ViewOptT vOpt;
    char *objName;
    ViewE vieType = VIEW_HEIGHT;
    int i;

    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
        return showcfgHeight(interp, NULL);
    }

    if (cmdAnalyzeName(argv[1], &objName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }
    
    if (dPtr->type != DSET_GRID && dPtr->type != DSET_SLICE) {
	Tcl_AppendResult(interp,
                         "heightfield is only valid for GRID or SLICE datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }
        
    hPtr = heiAlloc(argv[1]);
    if (!hPtr) {
	Tcl_AppendResult(interp, "couldn't create height", (char *) NULL);
	return TCL_ERROR;
    }
    hPtr->object = strDupl(objName);
    for (i=0; i<=1; ++i) {
        styFillFG(interp, &hPtr->style[i], "RGB 0.66 0.66 0.66");
        styFillPolygon(interp, &hPtr->style[i], "fill");
    }

    vOpt.doscale = 1;
    vOpt.doorig  = 0;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = -1.;
    vOpt.width = -1.;
    vOpt.height = -1.;
    if (configureHeight(interp, hPtr, argc-2, argv+2, &vOpt,
                        FALSE) != TCL_OK ) {
	goto Error;
    }

    vOpt.vrange[0] = dPtr->xsize;
    vOpt.vrange[1] = dPtr->ysize;
    if (cmdSetView(interp, &hPtr->view, 1, &vieType, objName,
                   argv[1], "height", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }

    Tcl_CreateCommand(interp, argv[1], HeightObjectCmd, (ClientData) hPtr,
		      (Tcl_CmdDeleteProc *) NULL );
    Tcl_SetVar2(interp, "active", "heightfield", argv[1], TCL_GLOBAL_ONLY);
    return TCL_OK;

  Error:
    heiFree(&hPtr);
    return TCL_ERROR;
}

int
HeightObjectCmd(ClientData data,		/* Height group	*/
		 Tcl_Interp *interp,		/* Current interpreter. */
		 int argc,			/* Number of arguments. */
		 char **argv)			/* Argument strings.	*/
{
    HeightP hPtr;
    ViewP view;
    ViewOptT vOpt;
    
    hPtr = (HeightP) data;
    view = viewFindByName(hPtr->view, NULL);
    
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"",
			 (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1],"configure",2)) {
        if (argc <= 2) {
            return showcfgHeight(interp, hPtr);
        } else {
            return configureHeight(interp, hPtr, argc-2, argv+2, &vOpt, FALSE);
        }
    } else if (strEQ(NULL,"clear",2)) {
        clearHeight(interp, view, hPtr);
    } else if (strEQ(NULL,"info",1)) {
	return infoHeight(interp, hPtr, argc-2, argv+2);
    } else if (strEQ(NULL,"plot",1)) {
        clearHeight(interp, view, hPtr);
	return plotHeight(interp, view, hPtr);
    } else if (strEQ(NULL,"remove",1)) {
        clearHeight(interp, view, hPtr);
        viewUnregisterObject(view, hPtr->name);
	Tcl_DeleteCommand(interp, hPtr->name);
	heiFree(&hPtr);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, clear, info, plot or remove",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;

}
    
static int
configureHeight(Tcl_Interp *interp,
                HeightP hPtr,
                int argc,
                char **argv,
                ViewOptT *vOpt,
                int append)
{
    DatasetP dPtr;
    GridP gPtr;
    int i, ival, has2styles;
    double value;
    int elemc;
    char **elemv;

    dPtr = setFindByName(hPtr->object);
    if (dPtr->type == DSET_GRID) {
        gPtr = &dPtr->data.dg->grid;
    } else {
        gPtr = &dPtr->data.ds->grid;
    }

    viewParseArgs(NULL, &argc, argv, vOpt);
    
    styParseArgs(&argc, argv, &hPtr->style[0], &hPtr->style[1], &has2styles);
    if (has2styles) {
        hPtr->flags |= M_HEI_HAS_COLORRAMP;
    } else {
        hPtr->flags &= ~ M_HEI_HAS_COLORRAMP;
    }
    
    for (i=0; i<argc; ++i) {
	if (strEQ(argv[i], "-cutoff", 2)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-cutoff\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
            if ( Tcl_SplitList(interp, argv[++i], &elemc, &elemv) != TCL_OK ) {
		return TCL_ERROR;
	    }
            if (elemc == 1) {
                if ( Tcl_GetDouble(interp, elemv[0], &value) != TCL_OK ) {
                    return TCL_ERROR;
                }
                hPtr->hcut = value;
                hPtr->lcut = -value;
            } else {
                if ( Tcl_GetDouble(interp, elemv[0], &hPtr->hcut) != TCL_OK ) {
                    return TCL_ERROR;
                }
                if ( Tcl_GetDouble(interp, elemv[1], &hPtr->lcut) != TCL_OK ) {
                    return TCL_ERROR;
                }
            }
        } else if (strEQ(NULL, "-map", 2)) {
            if (++i >= argc) {
                Tcl_AppendResult(interp, "value for \"-map\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            hPtr->map = strDuplOver(hPtr->map, argv[i]);
	} else if (strEQ(NULL, "-plane", 2)) {
	    if (++i >= argc) {
		Tcl_AppendResult(interp, "plane no. is missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[i], &ival) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    if (ival > gPtr->nz || ival < 1) {
		Tcl_AppendResult(interp, "plane no. outside valid range",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    hPtr->plane = ival-1;
        } else if (strEQ(NULL,"-vertac",3)) {
            if (++i >= argc) {
                Tcl_AppendResult(interp, "value for \"-vertac\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            if ( Tcl_GetDouble(interp, argv[i], &hPtr->vertac) != TCL_OK  ) {
                return TCL_ERROR;
            }
        } else if (strEQ(NULL,"-view",3)) {
            if (++i >= argc) {
                Tcl_AppendResult(interp, "value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            hPtr->view = strDupl(argv[i]);
	} else {
	    Tcl_AppendResult(interp, "unknown option \"",
			     argv[i], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

    }

    return TCL_OK;

}

static int
showcfgHeight(Tcl_Interp *interp,
              HeightP hPtr)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (hPtr) {
        Tcl_DStringAppendElement(&cfg, hPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "vertac");
    Tcl_DStringAppendElement(&cfg, "float");
/* Cljf    Tcl_DStringAppendElement(&cfg, "Vertical Acceleration");  */
    Tcl_DStringAppendElement(&cfg, "Vertical Accentuation"); 
    Tcl_DStringAppendElement(&cfg, "");
    if (hPtr) {
        sprintf(buf, "%g", hPtr->vertac);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "cutoff");
    Tcl_DStringAppendElement(&cfg, "float:2::high,low");
    Tcl_DStringAppendElement(&cfg, "Cutoff Values");
    Tcl_DStringAppendElement(&cfg, "");
    if (hPtr) {
        Tcl_DStringStartSublist(&cfg);
        sprintf(buf, "%g", hPtr->hcut);
        Tcl_DStringAppendElement(&cfg, buf);
        sprintf(buf, "%g", hPtr->lcut);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    /*
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "plane");
    Tcl_DStringAppendElement(&cfg, "int");
    Tcl_DStringAppendElement(&cfg, "Plane");
    Tcl_DStringAppendElement(&cfg, "");
    if (hPtr) {
        sprintf(buf, "%d", hPtr->plane);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "map");
    Tcl_DStringAppendElement(&cfg, "choice:1,sign,value,nothing");
    Tcl_DStringAppendElement(&cfg, "Map");
    Tcl_DStringAppendElement(&cfg, "nothing");
    if (hPtr && hPtr->map) {
        sprintf(buf, "%s", hPtr->map);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
    */
    if (hPtr) {
        ViewP view = viewFindByName(hPtr->view, NULL);
        styGetConfig(&cfg, &hPtr->style[0], STY_FG_ALL|STY_POLYGON_ALL,
                     0, (hPtr->flags & M_HEI_HAS_COLORRAMP), &hPtr->style[1]);
        if (view) {
            viewGetConfig(&cfg, &view->opt, 0);
        } else {
            viewGetConfig(&cfg, NULL, 0);
        }
    } else {
        styGetConfig(&cfg, NULL, STY_FG_ALL|STY_POLYGON_ALL, 0, 0, NULL);
        viewGetConfig(&cfg, NULL, 0);
    }

    if (hPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_flags");
        sprintf(buf, "%d", hPtr->flags);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
    
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_object");
        Tcl_DStringAppendElement(&cfg, hPtr->object);
        Tcl_DStringEndSublist(&cfg);
    }
        
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
infoHeight(Tcl_Interp *interp,
           HeightP hPtr,
           int argc,
           char **argv )
{
    Tcl_DString cmd;

    Tcl_DStringInit(&cmd);
    Tcl_DStringAppendElement(&cmd, "heiinfo");
    Tcl_DStringAppendElement(&cmd, hPtr->name);
    return Tcl_Eval(interp, Tcl_DStringValue(&cmd));
}

static int
plotHeight(Tcl_Interp *interp,
           ViewP view,
           HeightP hPtr)
{
    GridP gPtr;
    DatasetP dPtr;

    if (!view) return TCL_ERROR;

    if (hPtr->tag) return TCL_OK;
    
    dPtr = setFindByName(hPtr->object);
    if (dPtr->type == DSET_GRID) {
        gPtr = &dPtr->data.dg->grid;
    } else {
        gPtr = &dPtr->data.ds->grid;
    }
    hPtr->tag = dlAlloc(view, DL_HEIGHTFIELD, DL_MAIN_PLANE);
    dlMkHeight(hPtr->tag, hPtr->map, hPtr->plane, hPtr->vertac,
               hPtr->hcut, hPtr->lcut, gPtr, &hPtr->style[0]);
    pltUpdateView(view);
    return TCL_OK;

}

static int
clearHeight(Tcl_Interp *interp,
            ViewP view,
            HeightP hPtr)
{
    if (hPtr->tag) {
        dlRemove(view, hPtr->tag);
        hPtr->tag = 0;
    }
    return TCL_OK;
}
