#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/isoCommand.c,v 2.14 1997/03/29 18:06:24 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 *======================================================================
 *
 * Module: iso2DCmd
 *
 * Abstract: Defines the 'isosurface' command and creates isosurface object
 * commands.  Called from the tcl interpreter.
 *
 * $Log: isoCommand.c,v $
 * Revision 2.14  1997/03/29 18:06:24  richter
 * preset style
 *
 * Revision 2.13  1996/11/06  14:34:31  richter
 * (configureIsosurface): call clearIsosurface() before srfFree()
 *
 * Revision 2.12  1996/11/03  16:40:24  richter
 * styGetConfig expects a style group mask
 *
 * Revision 2.11  1996/09/06  11:20:58  richter
 * many changes to use the new display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:03:01  richter
 * use Tcl_SetVar2 to set active(isosurface)
 * handle 'iso <iso>' correctly
 * cmdSetView takes an array of possible view types
 *
 * Revision 2.9.1.2  1996/07/03  10:12:33  richter
 * refined view scaling
 * configure output
 *
 * Revision 2.9.1.1  1996/05/19  16:10:49  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:04  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "isoProto.h"
#include "cmdProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "setProto.h"
#include "strProto.h"
#include "styProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureIsosurface(Tcl_Interp *interp, IsosurfaceP cPtr,
			    int argc,char **argv, ViewOptT *vOpt, int append);
static int showcfgIsosurface(Tcl_Interp *interp, IsosurfaceP iPtr);
static int infoIsosurface(Tcl_Interp *interp, IsosurfaceP cPtr,
		       int argc,char **argv);
static int infoSurface(Tcl_Interp *interp, SurfaceP lPtr);
static int plotIsosurface(Tcl_Interp *interp, ViewP view, IsosurfaceP cPtr,
		       int argc,char **argv);
static int clearIsosurface(Tcl_Interp *interp, ViewP view, IsosurfaceP cPtr);
/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
IsosurfaceCmd(ClientData dummy,		/* Not used			*/
	   Tcl_Interp *interp,		/* Current interpreter		*/
	   int argc,			/* Number of arguments		*/
	   char **argv)			/* Argument strings		*/
{
    DatasetP dPtr;
    IsosurfaceP iPtr;
    ViewOptT vOpt;
    char *objName;
    ViewE vieType = VIEW_SURFACE;

    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
        return showcfgIsosurface(interp, NULL);
    }

    if (cmdAnalyzeName(argv[1], &objName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }

    if (dPtr->type != DSET_GRID && dPtr->type != DSET_SLICE) {
	Tcl_AppendResult(interp,
                         "isosurface is only valid for GRID or SLICE datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }
        
    iPtr = isoAlloc(argv[1]);
    if ( ! iPtr ) {
	Tcl_AppendResult(interp, "couldn't create isosurface", (char *) NULL);
	return TCL_ERROR;
    }
    iPtr->object = strDupl(objName);
    styFillFG(interp, &iPtr->style, "RGB 0.66 0.66 0.66");
    styFillPolygon(interp, &iPtr->style, "fill");
    
    vOpt.doscale = 1;
    vOpt.doorig  = 0;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = -1.;
    vOpt.width = -1.;
    vOpt.height = -1.;
    if (configureIsosurface(interp, iPtr, argc-2, argv+2,&vOpt,
                            FALSE) != TCL_OK ) {
	goto Error;
    }

    vOpt.vrange[0] = dPtr->xsize;
    vOpt.vrange[1] = dPtr->ysize;
    if (cmdSetView(interp, &iPtr->view, 1, &vieType, objName,
                   argv[1], "iso", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }
    
    Tcl_CreateCommand(interp, argv[1], IsosurfaceObjectCmd, (ClientData) iPtr,
		      (Tcl_CmdDeleteProc *) NULL );
    Tcl_SetVar2(interp, "active", "isosurface", argv[1], TCL_GLOBAL_ONLY);
    return TCL_OK;

  Error:
    isoFree(&iPtr);
    return TCL_ERROR;
}

int
IsosurfaceObjectCmd(ClientData data,		/* Isosurface group	*/
		 Tcl_Interp *interp,		/* Current interpreter. */
		 int argc,			/* Number of arguments. */
		 char **argv)			/* Argument strings.	*/
{
    IsosurfaceP iPtr;
    ViewP view;
    ViewOptT vOpt;
    
    iPtr = (IsosurfaceP) data;
    view = viewFindByName(iPtr->view, NULL);
    
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"",
			 (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1],"configure",2)) {
        if (argc <= 2) {
            return showcfgIsosurface(interp, iPtr);
        } else {
            return configureIsosurface(interp, iPtr, argc-2, argv+2, &vOpt,
                                       FALSE);
        }
    } else if (strEQ(NULL,"append",1)) {
	return configureIsosurface(interp, iPtr, argc-2, argv+2, &vOpt, TRUE);
    } else if (strEQ(NULL,"clear",2)) {
        clearIsosurface(interp, view, iPtr);
	srfClearPlotFlag(iPtr->surface);
    } else if (strEQ(NULL,"info",1)) {
	return infoIsosurface(interp, iPtr, argc-2, argv+2);
    } else if (strEQ(NULL,"plot",1)) {
        clearIsosurface(interp, view, iPtr);
	return plotIsosurface(interp, view, iPtr, argc-2, argv+2);
    } else if (strEQ(NULL,"remove",1)) {
        clearIsosurface(interp, view, iPtr);
        viewUnregisterObject(view, iPtr->name);
	Tcl_DeleteCommand(interp, iPtr->name);
	isoFree(&iPtr);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, append, plot or remove",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;

}
    
static int
configureIsosurface(Tcl_Interp *interp,
                    IsosurfaceP iPtr,
                    int argc,
                    char **argv,
                    ViewOptT *vOpt,
                    int append)
{
    DatasetP dPtr;
    GridP gPtr;
    SurfaceP *lpPtr;
    SurfaceP lp;
    int i, k;
    int ac, has2styles;
    double value;
    int recalc;
    int elemc;
    char **elemv;

    ViewP view;
    
    view = viewFindByName(iPtr->view, NULL);

    recalc = FALSE;

    dPtr = setFindByName(iPtr->object);
    if (dPtr->type == DSET_GRID) {
        gPtr = &dPtr->data.dg->grid;
    } else {
        gPtr = &dPtr->data.ds->grid;
    }

    viewParseArgs(NULL, &argc, argv, vOpt);
    ac = argc;
    styParseArgs(&argc, argv, &iPtr->style, NULL, &has2styles);
    if (ac != argc) {
        srfClearPlotFlag(iPtr->surface);
    }

    for (i=0; i<argc; ++i) {
        if (strEQ(argv[i],"-values",2)) {
	    i++;
	    if ( i >= argc ) {
		Tcl_AppendResult(interp, "no values given",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_SplitList( interp, argv[i], &elemc, &elemv ) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    lpPtr = &iPtr->surface;
	    lp = iPtr->surface;
	    if ( lp ) {
		if ( append ) {
		    while (lp->next) {
			lp = lp->next;
		    }
		    lpPtr = &lp->next;
		} else {
                    clearIsosurface(interp, view, iPtr);
		    srfFree(&lp);
		    iPtr->surface = NULL;
		    iPtr->ns =    0;
		}
	    }
	    for (k=0; k<elemc; ++k) {
		if ( Tcl_GetDouble(interp, elemv[k], &value) != TCL_OK ) {
		    goto Error;
		}
		if ( ! (lp = (*lpPtr) = srfAlloc()) ) {
		    Tcl_AppendResult(interp, "couldn't create surface",
				     "(at ", elemv[k], ")", (char *) NULL);
		    ckfree((char *) elemv);
		    return TCL_ERROR;
		}
		lp->value = value;
		lpPtr = &lp->next;
		iPtr->ns += 1;
	    }
	    ckfree((char *) elemv);
	    recalc = TRUE;
        } else if (strEQ(NULL,"-view",3)) {
            i++;
            if (i >= argc) {
                Tcl_AppendResult(interp, "value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            iPtr->view = strDupl(argv[i]);
	} else {
	    Tcl_AppendResult(interp, "unknown option \"",
			     argv[i], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (recalc) {
	isoEngine(gPtr, iPtr);
    }
    return TCL_OK;

Error:
    ckfree((char *) elemv);
    elemv = NULL;
    return TCL_ERROR;
}

static int
showcfgIsosurface(Tcl_Interp *interp,
                  IsosurfaceP iPtr)
{
    SurfaceP lp;
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (iPtr) {
        Tcl_DStringAppendElement(&cfg, iPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
    
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "values");
    Tcl_DStringAppendElement(&cfg, "string::required");
    Tcl_DStringAppendElement(&cfg, "Surface Values");
    Tcl_DStringAppendElement(&cfg, "");
    if (iPtr) {
        Tcl_DStringStartSublist(&cfg);
        for(lp=iPtr->surface; lp; lp=lp->next) {
            sprintf(buf, "%g", lp->value);
            Tcl_DStringAppendElement(&cfg, buf);
            Tcl_DStringEndSublist(&cfg);
        }
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    if (iPtr) {
        ViewP view = viewFindByName(iPtr->view, NULL);
        styGetConfig(&cfg, &iPtr->style, STY_FG_ALL|STY_POLYGON_ALL,
                     0, 0, NULL);
        if (view) {
            viewGetConfig(&cfg, &view->opt, 0);
        } else {
            viewGetConfig(&cfg, NULL, 0);
        }
    } else {
        styGetConfig(&cfg, NULL, STY_FG_ALL|STY_POLYGON_ALL, 0, 0, NULL);
        viewGetConfig(&cfg, NULL, 0);
    }

    if (iPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_flags");
        sprintf(buf, "%d", iPtr->flags);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
        
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_object");
        Tcl_DStringAppendElement(&cfg, iPtr->object);
        Tcl_DStringEndSublist(&cfg);
    }
    
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
infoIsosurface(Tcl_Interp *interp,
               IsosurfaceP iPtr,
               int argc,
               char **argv )
{
    sprintf(buf, "Isosurface '%s'  Flags: 0x%2.2x\n",
	   iPtr->name, iPtr->flags);
    Tcl_AppendResult(interp, buf, NIL);
    sprintf(buf, "\tView: '%s'\n", iPtr->view);
    Tcl_AppendResult(interp, buf, NIL);
    sprintf(buf, "\tObject:  '%s'\n", iPtr->object);
    Tcl_AppendResult(interp, buf, NIL);
    sprintf(buf, "\t%d surfaces\n", iPtr->ns);
    Tcl_AppendResult(interp, buf, NIL);
    return infoSurface(interp, iPtr->surface);
}

static int
infoSurface(Tcl_Interp *interp,
            SurfaceP lPtr)
{
    if (lPtr == NULL) {
	return TCL_OK;
    }
    sprintf(buf, "\t %g 0x%2.2x %d\n", lPtr->value, lPtr->flags, lPtr->nt);
    Tcl_AppendResult(interp, buf, NIL);
    return infoSurface(interp, lPtr->next);
}

static int
plotIsosurface(Tcl_Interp *interp,
               ViewP view,
               IsosurfaceP iPtr,
               int argc,
               char **argv )
{
    int force;
    int i;
    size_t length;
    SurfaceP lPtr;

    if (!view) return TCL_ERROR;

    if (argc > 1 ) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			 " plot ?-force?\"", NIL);
	return TCL_ERROR;
    } else if (argc == 1 && (length = strlen(argv[0])) > 0) {
	if ( strncmp(argv[0],"-force",length) == 0 ) {
	    force = TRUE;
	} else {
	    Tcl_AppendResult(interp, "bad option \"", argv[0],
			     "\": must be -force", NIL);
	    return TCL_ERROR;
	}
    } else {
	force = FALSE;
    }

    for (i=0, lPtr = iPtr->surface; lPtr; ++i, lPtr = lPtr->next) {
        if (lPtr->tag) continue;
	if ( !force && (lPtr->flags & M_SRF_PLOT_DONE)) {
	    continue;
	}
#ifdef DEBUG
        pltMessage("inf", "surface %8.3f  ", lPtr->value);
#endif
        lPtr->tag = dlAlloc(view, DL_TRIANGLE, DL_MAIN_PLANE);
        dlMkTriangle(lPtr->tag, lPtr->triangles, &iPtr->style);
	lPtr->flags |= M_SRF_PLOT_DONE;
    }
    pltUpdateView(view);
    return TCL_OK;
}

static int
clearIsosurface(Tcl_Interp *interp,
                ViewP view,
                IsosurfaceP iPtr)
{
    int i;
    SurfaceP lPtr;
        
    for (i=0, lPtr = iPtr->surface; lPtr; ++i, lPtr = lPtr->next) {
        lPtr->flags &= ~ M_SRF_PLOT_DONE;
        if (lPtr->tag != 0) {
            dlRemove(view, lPtr->tag);
            lPtr->tag = 0;
        }
    }
    return TCL_OK;
}
