#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/isoEngine.c,v 2.14 1997/03/31 15:21:36 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 *======================================================================
 *
 * Module:
 *
 *
 * $Log: isoEngine.c,v $
 * Revision 2.14  1997/03/31 15:21:36  richter
 * added some blanks to remove =* ambiguities
 *
 * Revision 2.13  1996/12/20  15:03:27  richter
 * added some prototypes
 *
 * Revision 2.12  1996/12/14  12:54:30  richter
 * (generate_coords): fixed a bug which did not show up in cubic volumes
 *
 * Revision 2.11  1996/11/03  16:34:17  richter
 * fixed some returns for void routines
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/03  11:00:52  richter
 * de-linting
 *
 * Revision 2.9.1.1  1996/05/20  11:30:54  richter
 * remove some debug output
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:04  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"            /* type definitions and array sizes     */
#include "isoProto.h"

#include "isoPattern.h"
/*------------------------------------------------------------------*/

#define PATNO 256
#define BOUND(x,t) (x>=t ? 1 : 0)
#define EXTRACT(x,n) (x>>(n*3))&07
#define EXTRACT1(x,n) (x>>n)&01
/*-------------------------------ba9 876 543 210 */
#define XTRACED 01122         /* 001 001 010 010 */
#define YTRACED 00461         /* 000 100 110 001 */
#define ZTRACED 00017         /* 000 000 001 111 */
#define PlyNo1 1
#define PlyNo2 2
#define PlyNo3 2
#define PlyNo4 2
#define PlyNo5 3
#define PlyNo6 2
#define PlyNo7 3
#define PlyNo8 2
#define PlyNo9 4
#define PlyNo10 4
#define PlyNo11 4
#define PlyNo12 4
#define PlyNo13 4
#define PlyNo14 4

/*------------------------------------------------------------------*/
/* one pattern table with 128 entries is defined */
static unsigned int pattab[PATNO];

/*------------------------------------------------------------------*/
typedef struct INTERINFO *IIP;
typedef struct INTERINFO {
   double pos[3];
   double norm[3];
} IIT;


/* STATIC GLOBAL VARIABLES */
static double *xp,*yp,*zp,*coord;
static struct INTERINFO *b,*f,*m,*tbf;
static int pno[15];
static int xtraced, ytraced, ztraced;
static int dim0, dim1, dim2, nvplan, nvcube;


typedef void (Add_Triangle) (void *clientdata,
                             double *v1, double *v2, double *v3,
                             double *n1, double *n2, double *n3);

static void create_isosurf(GridP gPtr, double level,
                           Add_Triangle *addtriangle, void *clientdata);

static void read1plane(Add_Triangle *addtriangle, void *clientdata,
                       double *d, double level, int zoff, int z, int *size);
static void copy_nv(IIP bmf, double *v, double *n);
static void copy_bmf(IIP bmf, double *v, double *n);
static unsigned int getpcode(double *d, int z0, int y0, int x, double level);
static void generate_coords(GridP gPtr, double *coord);
static void find_offset(double *d, unsigned char c,
                        int z0, int y0, int x, int y, int z,
                        double *v, double *n, double *vd);
static void interpolate(double *v1, double *v2, double d1, double d2,
                        double level, double *vert);
static void init_pattab(void);

void
isoEngine(GridP gPtr,
          IsosurfaceP iPtr)
{
    SurfaceP surface;
    
   /* initialize pattern table as well-defined pattern */
   init_pattab();

   /* process the input and create the isosurface geom */
   surface = iPtr->surface;
   while (surface) {
       create_isosurf(gPtr, surface->value, srfAddTriangle, (void *) surface);
       surface->flags |= M_SRF_CALC_DONE;
       surface = surface->next;
    }

   return;
}

/* Main routine to generate isosurface geometric object by Marching method
   Note: the input field is a uniform one ???
*/
/*------------------------------------------------------------------*/
static void
create_isosurf(GridP gPtr,
               double level,
               Add_Triangle *addtriangle,
               void *clientdata)
{

/*------------------------------------------------------------------*/
  /* pointer to input data */
  double *data=gPtr->va;
  int size;
  /* miscellaneous local variable declarations */
  int zoff0, zoff1;
  int i,k;
/*------------------------------------------------------------------*/
  IIP backp, frontp, middlp;
  
  /* get dimension for x, y and z and pre-compute multiply */
  dim0=gPtr->nx;
  dim1=gPtr->ny;
  dim2=gPtr->nz;
  nvplan=dim0*dim1;
  nvcube=nvplan*dim2;

  /* generate uniform coordinate list */
  coord=(double *)malloc(3*nvcube * sizeof(double));
  if (coord==NULL) {
      fprintf(stderr,
              "Malloc failure in isosurf_p for coordinates.");
      exit(1);
  }
  generate_coords(gPtr, coord);

  backp=(IIP) malloc((2*nvplan-dim0-dim1) * sizeof(IIT));
  if (backp==NULL) {
      fprintf(stderr,
              "Malloc failure in isosurf_p for back plane coordinates.");
      exit(1);
  }

  frontp=(IIP) malloc((2*nvplan-dim0-dim1) * sizeof(IIT));
  if (frontp==NULL) {
      fprintf(stderr,
              "Malloc failure in isosurf_p for front plane coordinates.");
      exit(1);
  }

  middlp=(IIP) malloc(nvplan * sizeof(IIT));
  if (middlp==NULL) {
      fprintf(stderr,
              "Malloc failure in isosurf_p for middle plane coordinates.");
      exit(1);
  }

  /* xp -->start of all x coord under dimension */
  xp=coord;
  /* yp -->start of all y coord under dimension */
  yp=xp+nvcube;
  /* zp -->start of all z coord under dimension */
  zp=yp+nvcube;

  for (i=0;i<15;i++) pno[i]=0;

  /* for all cubes match pattern code table to generate triangles */
  zoff1=0;
  b=backp;
  m=middlp;
  f=frontp;
  for (k=0,ztraced=0;k<dim2-1;k++,ztraced=ZTRACED) {
      zoff0=zoff1;
      zoff1=zoff0+nvplan;
      /* deal with only one z plane cubes */
      read1plane(addtriangle, clientdata, data, level, zoff0, k, &size);
      /* redirect front pointer to back pointer */
      tbf=f;
      f=b;
      b=tbf;

  }

  /* for (i=0;i<15;i++) printf("Type %d = %d\n",i,pno[i]); /* */
  /* free all tempory memory */
     free(backp);
     free(middlp);
     free(frontp);
     free(coord);

  return;
}

/* read 1st cubic plane with positions and normals */
static void
read1plane(Add_Triangle *addtriangle,
           void *clientdata,
           double *d,
           double level,
           int zoff,
           int z,
           int *size)
{
   int i,j,k;
   double vdata1,vdata2;
   unsigned int pcode;
   unsigned int ptype;
   unsigned int cedge;
   unsigned int uedge;
   unsigned int c1,c2;
   double *addressv,*addressn;
   struct INTERINFO edge_array[12];
   char *plyedges;
   char PlyNo;
   char edge_cnvt[12];
   char e1,e2,e3;
   double v0[3],v1[3];
   double n0[3],n1[3];
   int yoff1,yoff0,zoff1,zoff0;

   /*
    * z & y offset initialization
    */
   zoff0=zoff;
   zoff1=zoff0+nvplan;
   yoff1=0;
   *size=0;

   /*
    * begin to find isosurface
    */
   /* loop for y direction with yExt=dim1-1 */
   for (i=0,ytraced=ztraced;i<dim1-1;i++,ytraced=ztraced,ytraced|=YTRACED) {
       yoff0=yoff1;
       yoff1=yoff0+dim0;
       /*
        * loop for x direction with xExt=dim0-1
        */
       for (j=0,xtraced=ytraced;j<dim0-1;j++,xtraced=ytraced,xtraced|=XTRACED) {
           /* get pattern code from the current cubic */
           pcode=getpcode(d,zoff0,yoff0,j,level);
           ptype=pcode;
           ptype=(ptype>>24)&077;
           pno[ptype]++;
           /* this case will not happen */
           if (ptype>14) {
              fprintf(stderr,"Error pattern type !!");
              exit(1);
           }

           /* get cross edge code for ptype */
           cedge=CROSSEDGE[ptype];

           /*
            * calculate intersected position
            */
           for (k=0;k<12;cedge>>=1,k++) {
               /* only consider the edges that are intersected by isosurface */
               if (cedge&1) {

                  /* convert real corner index for these two end points */
                  c1=EXTRACT(pcode,UEDGETABLE[k][0]);
                  c2=EXTRACT(pcode,UEDGETABLE[k][1]);

                  /* decide real edge that is intersected */
                  uedge=EDGETABLE[c1][c2];

                  /* ready for retrieval later */
                  edge_cnvt[k]=uedge;

                  /* redirect edge_array */
                  addressv=edge_array[uedge].pos;
                  addressn=edge_array[uedge].norm;

                  if (EXTRACT1(xtraced,uedge)) {
                     /* this is a shared edge and only get data from cache */
                     switch (uedge) {
                     case  0:
                         copy_nv(f+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  1:
                         copy_nv(f+dim0-1+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  2:
                         copy_nv(f+dim0-1+yoff0+yoff0-i+j+1,addressv,addressn);
                         break;
                     case  3:
                         copy_nv(f+yoff0+dim0+yoff0+dim0-i-1+j,addressv,addressn);
                         break;
                     case  4:
                         copy_nv(m+yoff0+j,addressv,addressn);
                         break;
                     case  5:
                         copy_nv(m+yoff0+j+1,addressv,addressn);
                         break;
                     case  6:
                         copy_nv(m+yoff1+j,addressv,addressn);
                         break;
                     case  8:
                         copy_nv(b+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  9:
                         copy_nv(b+dim0-1+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  7:
                     case 10:
                     case 11:
                     default: break;
                     }
                  } else {
                  /* v,n must be generated and copy them onto bmf */
                     find_offset(d,c1,zoff0,yoff0,j,i,z,v0,n0,&vdata1);
                     find_offset(d,c2,zoff0,yoff0,j,i,z,v1,n1,&vdata2);

                     /* here v0,n0,vdata1 & v1,n1,vdata2 are available */
                     /* interpolate these two end points into the edge */
                     interpolate(v0,v1,vdata1,vdata2,level,addressv);
                     interpolate(n0,n1,vdata1,vdata2,level,addressn);

                     /* record it into cache */
                     switch (uedge) {
                     case  0:
                         copy_bmf(f+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  1:
                         copy_bmf(f+dim0-1+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  2:
                         copy_bmf(f+dim0-1+yoff0+yoff0-i+j+1,addressv,addressn);
                         break;
                     case  3:
                         copy_bmf(f+yoff0+dim0+yoff0+dim0-i-1+j,addressv,addressn);
                         break;
                     case  4:
                         copy_bmf(m+yoff0+j,addressv,addressn);
                         break;
                     case  5:
                         copy_bmf(m+yoff0+j+1,addressv,addressn);
                         break;
                     case  6:
                         copy_bmf(m+yoff1+j,addressv,addressn);
                         break;
                     case  8:
                         copy_bmf(b+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  9:
                         copy_bmf(b+dim0-1+yoff0+yoff0-i+j,addressv,addressn);
                         break;
                     case  7:
                         copy_bmf(m+yoff1+j+1,addressv,addressn);
                         break;
                     case 10:
                         copy_bmf(b+dim0-1+yoff0+yoff0-i+j+1,addressv,addressn);
                         break;
                     case 11:
                         copy_bmf(b+yoff0+dim0+yoff0+dim0-i-1+j,addressv,addressn);
                         break;
                     default: break;
                     }
                  
                  /* assign address pointer to be ready for fill coords and norms
                                   (8)
                              4------------5
                        (4)  /|       (5) /|
                             0------------1|
                             ||    (0)    ||        edge index assignment (0..11)
                             ||(9)        ||(10)
                        (1)  ||       (2) ||        corner index assignment (0..7)
                             |6___(11)____|7
                             2/(6)_ ______3/(7)
                                   (3)
                  */
                  }

               }
               /* unsetting edge doesn't care !! */
           }
           /* complete all these 12 edge checkings */
           /* PlyNo     : decides how many triangles to generate
            * plyedges  : defines all triangle edge index
            * both variables must be pointered to appropriate data
            * structure for different type
            */
/*-----------------------------------------------------------*/
           switch (ptype) {
           case 1:
               PlyNo=PlyNo1;                     
               plyedges= &PLYEDGES1[0][0];
               break;
           case 2:
               PlyNo=PlyNo2;
               plyedges= &PLYEDGES2[0][0];
               break;
           case 3:
               PlyNo=PlyNo3;
               plyedges= &PLYEDGES3[0][0];
               break;
           case 4:
               PlyNo=PlyNo4;
               plyedges= &PLYEDGES4[0][0];
               break;
           case 5:
               PlyNo=PlyNo5;
               plyedges= &PLYEDGES5[0][0];
               break;
           case 6:
               PlyNo=PlyNo6;
               plyedges= &PLYEDGES6[0][0];
               break;
           case 7:
               PlyNo=PlyNo7;
               plyedges= &PLYEDGES7[0][0];
               break;
           case 8:
               PlyNo=PlyNo8;
               plyedges= &PLYEDGES8[0][0];
               break;
           case 9:
               PlyNo=PlyNo9;
               plyedges= &PLYEDGES9[0][0];
               break;
           case 10:
               PlyNo=PlyNo10;
               plyedges= &PLYEDGES10[0][0];
               break;
           case 11:
               PlyNo=PlyNo11;
               plyedges= &PLYEDGES11[0][0];
               break;
           case 12:
               PlyNo=PlyNo12;
               plyedges= &PLYEDGES12[0][0];
               break;
           case 13:
               PlyNo=PlyNo13;
               plyedges= &PLYEDGES13[0][0];
               break;
           case 14:
               PlyNo=PlyNo14;
               plyedges= &PLYEDGES14[0][0];
               break;
           default:
               PlyNo=0;
               break;
           }
/*------------------------------------------------------------*/
           /*
            * here actually generate geometry object for triangles
            */
           for (k=0;k<PlyNo;k++) {
               /* convert to correct edge lists */
               e1=edge_cnvt[*(plyedges+k*3)];
               e2=edge_cnvt[*(plyedges+k*3+1)];
               e3=edge_cnvt[*(plyedges+k*3+2)];
               
               addtriangle(clientdata, edge_array[e1].pos, edge_array[e2].pos,
                           edge_array[e3].pos,
                           edge_array[e1].norm, edge_array[e2].norm,
                           edge_array[e3].norm);
           }
           *size= *size + PlyNo;
        }
        /* end of j */
     }
     /* end of i */
}

/* copy from bmf to n,v */
static void
copy_nv(IIP bmf,
        double *v,
        double *n)
{
   *v=bmf->pos[0];  *(v+1)=bmf->pos[1];  *(v+2)=bmf->pos[2];
   *n=bmf->norm[0]; *(n+1)=bmf->norm[1]; *(n+2)=bmf->norm[2];
   return;
}

/* copy from n,v to bmf */
static void
copy_bmf(IIP bmf,
         double *v,
         double *n)
{
   bmf->pos[0]= *v;  bmf->pos[1]= *(v+1);  bmf->pos[2]= *(v+2);
   bmf->norm[0]= *n; bmf->norm[1]= *(n+1); bmf->norm[2]= *(n+2);
   return;
}

/* get vert,norm and data for interesting cubic index */
static void
find_offset(double *d,
            unsigned char c,
            int z0, int y0, int x, int y, int z,
            double *v, double *n,
            double *vd)
{
/*------------------------------------------------------------*/
    int offset;
    int z1=z0+nvplan;
    int y1=y0+dim0;
    double Gx,Gy,Gz,G,H;
    int X=x;
    int Y=y;
    int Z=z;
/*------------------------------------------------------------*/

/* adjust X,Y and Z and the offset to get data from */
    switch (c) {
    case 0:
        offset=z0+y0+x;
        break;
    case 1:
        offset=z0+y0+x+1;
        X++;
        break;
    case 2:
        offset=z0+y1+x;
        Y++;
        break;
    case 3:
        offset=z0+y1+x+1;
        X++; Y++;
        break;
    case 4:
        offset=z1+y0+x;
        Z++;
        break;
    case 5:
        offset=z1+y0+x+1;
        Z++; X++;
        break;
    case 6:
        offset=z1+y1+x;
        Z++; Y++;
        break;
    case 7:
        offset=z1+y1+x+1;
        Z++; Y++; X++;
        break;
    default:
        break;
    }

    /* get real coordinates for vertex v */
    *v= *(xp+offset);
    *(v+1)= *(yp+offset);
    *(v+2)= *(zp+offset);

  /* get real normals for normal n */
    if (X==0 || X==(dim0-1)) {
        *n = 0.0;
    } else {
        *n = *(d+offset+1) - *(d+offset-1);
    }
    if (Y==0 || Y==(dim1-1)) {
        *(n+1) = 0.0;
    } else {
        *(n+1) = *(d+offset+dim0) - *(d+offset-dim0);
    }
    if (Z==0 || Z==(dim2-1)) {
        *(n+2) = 0.0;
    } else {
        *(n+2) = *(d+offset+nvplan) - *(d+offset-nvplan);
    }
    Gx = - *n;
    Gy = - *(n+1);
    Gz = - *(n+2);

    G=Gx*Gx+Gy*Gy+Gz*Gz;
    if (G>0.0) {
        H=sqrt(G);
        *n=Gx/H;
        *(n+1)=Gy/H;
        *(n+2)=Gz/H;
    }

  /* get vertice data */
    *vd= *(d+offset);
    return;
}

/* interpolate data for coord */
static void
interpolate(double *v1, double *v2,
            double d1, double d2,
            double level,
            double *vert)
{
   double u;
   u=(level-d1)/(d2-d1);
   *vert= *v1+u*(*v2-*v1);
   *(vert+1)= *(v1+1)+u*(*(v2+1)-*(v1+1));
   *(vert+2)= *(v1+2)+u*(*(v2+2)-*(v1+2));
   return;
}
/* get pattern code for some cube
   the eight pos for this cube is arranged as the following :


       z1+y0+x      z1+y0+x+1
       (4)X=============X(5)
         /|            /|
     (0)X=============X(1)
     z0+y0+x       z0+y0+x+1     ======>  bit 7 6 5 4 3 2 1 0
        | |           | |                 -------------------
        | |(6)        | |(7)                  0 0 1 1 0 1 0 0
        |/z1+y1+x=====|/z1+y1+x+1             LOOK UP PATTAB
     (2)X=============X(3)                          ||
     z0+y1+x       z0+y1+x+1             (int)(6 bits)(8X3 bits)
                                                 |        |
                                           pattern type   |
                                             result after rotations

*/
static unsigned int
getpcode(double *d,
         int z0,
         int y0,
         int x,
         double level)
{
   unsigned char code=0,tcode,sum1=0;
   int z1=z0+nvplan;
   int y1=y0+dim0;
   /* 7 */
   tcode=BOUND((*(d+z1+y1+x+1)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 6 */
   tcode=BOUND((*(d+z1+y1+x)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 5 */
   tcode=BOUND((*(d+z1+y0+x+1)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 4 */
   tcode=BOUND((*(d+z1+y0+x)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 3 */
   tcode=BOUND((*(d+z0+y1+x+1)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 2 */
   tcode=BOUND((*(d+z0+y1+x)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 1 */
   tcode=BOUND((*(d+z0+y0+x+1)),level);
   code|=tcode; code<<=1; sum1+=tcode;
   /* 0 */
   tcode=BOUND((*(d+z0+y0+x)),level);
   code|=tcode; sum1+=tcode;
   /* here sum1 accumulates # of 1 in code */

   /* if # of 1 greater than four then inverse it */
   if (sum1>4)
      code=~code;

   /* return code in the pattern table */
   return(pattab[code]);

}

static void
generate_coords(GridP gPtr,
                double *coord)
{
  int i, j, k;
  double *x, *y, *z;

  /* pointer to coord structure for x,y and z */
  x=coord;
  y=x+nvcube;
  z=y+nvcube;

  for (k=0; k < gPtr->nz; k++){
      for (j=0; j < gPtr->ny; j++){
          for (i=0; i < gPtr->nx; i++){
              *x++ = i*gPtr->xinc - gPtr->xran/2.;
              *y++ = j*gPtr->yinc - gPtr->yran/2.;
              *z++ = k*gPtr->zinc - gPtr->zran/2.;
          }
      }
  }
  return;
}


static void
init_pattab(void)
{
   int i;

   /* set initial pattern value */
   for (i=0;i<=PATNO-1;i++) {
       pattab[i]=01700000000;  /* only for checking empty pattern use */
   }
   /* fill special pattern values from predefined ones */
   for (i=0;i<Pattern0No;i++)
       pattab[PAT0[i].cubeindex]=PAT0[i].patcode;

   for (i=0;i<Pattern1No;i++)
       pattab[PAT1[i].cubeindex]=PAT1[i].patcode;

   for (i=0;i<Pattern2No;i++)
       pattab[PAT2[i].cubeindex]=PAT2[i].patcode;

   for (i=0;i<Pattern3No;i++)
       pattab[PAT3[i].cubeindex]=PAT3[i].patcode;

   for (i=0;i<Pattern4No;i++)
       pattab[PAT4[i].cubeindex]=PAT4[i].patcode;

   for (i=0;i<Pattern5No;i++)
       pattab[PAT5[i].cubeindex]=PAT5[i].patcode;

   for (i=0;i<Pattern6No;i++)
       pattab[PAT6[i].cubeindex]=PAT6[i].patcode;

   for (i=0;i<Pattern7No;i++)
       pattab[PAT7[i].cubeindex]=PAT7[i].patcode;

   for (i=0;i<Pattern8No;i++)
       pattab[PAT8[i].cubeindex]=PAT8[i].patcode;

   for (i=0;i<Pattern9No;i++)
       pattab[PAT9[i].cubeindex]=PAT9[i].patcode;

   for (i=0;i<Pattern10No;i++)
       pattab[PAT10[i].cubeindex]=PAT10[i].patcode;

   for (i=0;i<Pattern11No;i++)
       pattab[PAT11[i].cubeindex]=PAT11[i].patcode;

   for (i=0;i<Pattern12No;i++)
       pattab[PAT12[i].cubeindex]=PAT12[i].patcode;

   for (i=0;i<Pattern13No;i++)
       pattab[PAT13[i].cubeindex]=PAT13[i].patcode;

   for (i=0;i<Pattern14No;i++)
       pattab[PAT14[i].cubeindex]=PAT14[i].patcode;

   return;
}
