#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/namManager.c,v 2.10 1996/07/25 09:14:52 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: namManager.c,v $
 * Revision 2.10  1996/07/25 09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.1  1996/07/03  14:30:48  richter
 * paranoia fix
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:05  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "namProto.h"

/*
 * variables with internal linkage
 */
static Tcl_HashTable namHashTable;
static int namHashTable_unused = 1;

int
namAlloc(char *type,
         char *name,
         void *data)
{
    Tcl_HashEntry *hPtr;
    int new;
    char *key;

    key = ckalloc(strlen(type)+strlen(name)+2);
    key = strcpy(key, type);
    key = strcat(key, ":");
    key = strcat(key, name);
    
    if (namHashTable_unused) {
	Tcl_InitHashTable(&namHashTable, TCL_STRING_KEYS);
	namHashTable_unused = 0;
    }
    hPtr = Tcl_CreateHashEntry(&namHashTable, key, &new);
    if (! new) {
    	Tcl_AppendResult(interp, "name \"", name, "\" already exists", NIL);
	return TCL_ERROR;
    }

    Tcl_SetHashValue(hPtr, (ClientData) data);
    return TCL_OK;
}

void
namFree(char *type,
        char *name)
{
    Tcl_HashEntry *hPtr;
    char *key;

    key = ckalloc(strlen(type)+strlen(name)+2);
    key = strcpy(key, type);
    key = strcat(key, ":");
    key = strcat(key, name);
    hPtr = Tcl_FindHashEntry(&namHashTable, key);
    if (hPtr) {
	Tcl_DeleteHashEntry(hPtr);
    }
    ckfree(key);
    return;
}

int
namListByName(char *type,
              char *match)
{
    char *name;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    char *key, *colon;

    key = ckalloc(strlen(type)+strlen(match)+2);
    key = strcpy(key, type);
    key = strcat(key, ":");
    key = strcat(key, match);

    for (hPtr = Tcl_FirstHashEntry(&namHashTable, &search);
	 hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
	name = Tcl_GetHashKey(&namHashTable, hPtr);
	if (Tcl_StringMatch(name, key)) {
            colon = strchr(name,':');
            if (colon) {
                name = colon+1;
            }
	    Tcl_AppendElement(interp, name);
	}
    }
    ckfree(key);
    return TCL_OK;
}

void *
namFindByName(char *type,
              char *name)
{
    Tcl_HashEntry *hPtr;
    char *key;

    if (namHashTable_unused || !name || !type) {
        return (void *) NULL;
    }
    key = ckalloc(strlen(type)+strlen(name)+2);
    key = strcpy(key, type);
    key = strcat(key, ":");
    key = strcat(key, name);
    hPtr = Tcl_FindHashEntry(&namHashTable, key);
    ckfree(key);
    
    if (hPtr) {
	return (void *) Tcl_GetHashValue(hPtr);
    } else {
	return (void *) NULL;
    }
}
