#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/pltInit.c,v 2.14 1997/03/29 18:09:55 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: pltInit.c,v $
 * Revision 2.14  1997/03/29 18:09:55  richter
 * new parameter 'what' for tkFlush
 *
 * Revision 2.13  1996/09/14  15:13:42  richter
 * return error code if styLoadElements fails
 *
 * Revision 2.12  1996/09/06  17:52:30  richter
 * moevit and drawit return an error status
 *
 * Revision 2.11  1996/09/06  11:27:22  richter
 * new routine pltGetMarker
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/03  10:11:53  richter
 * read element file
 *
 * Revision 2.9.1.1  1996/05/19  16:09:49  richter
 * some changes caused by transformation/view cleanup
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:09  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: pltDispatch
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration			*/
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "../../libxd/src/getline.h"	/* getline interface		*/
#define getCommand getline
#include "drvProto.h"
#include "pltProto.h"
#include "styProto.h"
#include "strProto.h"
#include "pltMarker.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static void pltCommandLoop(void);

/*----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */
TheDriver driver = {NODRIVER, 0};
int (*drawit)(void *,double,double);
int (*moveit)(void *,double,double);
MarkDataT *Marker[NUM_MARKER];

int
pltInit(char *d, char *init)
{
    int n, length, code;
    Tcl_DString rgbtxt, defmap, elemfile;
    MarkDataT *mPtr;
#ifdef WIN32          
    void xdMessageBox (char *string);
#endif
    /*
     * Driver-independent initializations must be done before we call
     * specific initialization routines, which might contain driver
     * specific command loops.
     */
    Tcl_DStringInit(&rgbtxt);
    Tcl_DStringAppend(&rgbtxt, Tcl_GetVar(interp, "xd_datadir", TCL_GLOBAL_ONLY), -1);
    Tcl_DStringAppend(&rgbtxt, "/rgb.txt", -1);
    styLoadColorname(Tcl_DStringValue(&rgbtxt));
    Tcl_DStringFree(&rgbtxt);
 
    Tcl_DStringInit(&defmap);
    Tcl_DStringAppend(&defmap, Tcl_GetVar(interp, "xd_datadir", TCL_GLOBAL_ONLY), -1);
    Tcl_DStringAppend(&defmap, "/default.cmap", -1);
    styLoadColormap(0, Tcl_DStringValue(&defmap));
    Tcl_DStringFree(&defmap);

    Tcl_DStringInit(&elemfile);
    Tcl_DStringAppend(&elemfile, Tcl_GetVar(interp, "xd_datadir", TCL_GLOBAL_ONLY), -1);
    Tcl_DStringAppend(&elemfile, "/elements.style", -1);
    code = styLoadElements(interp, Tcl_DStringValue(&elemfile));
    Tcl_DStringFree(&elemfile);
    if (code != TCL_OK) return code;
    
    for (n=0; n < NUM_MARKER; ++n) {
        Marker[n] = (MarkDataT *) 0;
    }
    for (mPtr = &mark_data[0], n=0;
         mPtr->mode > -99 && n < NUM_MARKER;
         ++mPtr) {
        if (mPtr->mode < 0) {
            Marker[n++] = mPtr;
        }
    }
    length = strlen(d);
    if ( strncmp(d, "cc", length) == 0 ) {
#ifdef CC_OUTPUT
        drawit = ccdraw;
        moveit = ccmove;
	driver.type = CC;
        driver.transforms = 0;
	plots(20., 20., 61);
#else

#ifdef WIN32
        xdMessageBox("CalComp driver not installed with this version");
#else
	fprintf(stderr,"calcomp driver not installed\n");
#endif
	exit(1);
#endif

    } else if ( strncmp(d, "tk", length) == 0 ) {

#ifdef TK_OUTPUT     
        drawit = tkDraw;
        moveit = tkMove;
	driver.type = TK;
        driver.transforms = 0;
	tkInit(init);
        /* NOT REACHED */
#else
#ifdef WIN32
        xdMessageBox("Tk driver not installed with this version");
#else
	fprintf(stderr,"tk driver not installed\n");
#endif
	exit(1);
#endif
    } else if ( strncmp(d, "gt", length) == 0 ) {
#ifdef GT_OUTPUT
	driver.type = GT;
        driver.transforms = 1;
	gtInit(init);
#else

#ifdef WIN32
        xdMessageBox("Open GL/Tk driver not installed with this version");
#else
	fprintf(stderr,"open gl/tk driver not installed\n");
#endif
	exit(1);
#endif

    } else {

#ifdef WIN32
        xdMessageBox("Unknown driver requested. Should be tk or gt");
#else
	fprintf(stderr,"unknown driver \"%s\": should be tk, gt or cc", d);
#endif
	exit(1);
    }

    /*
     * This is only used for drivers without their own command loop.
     * Currently CALCOMP only.
     */
    pltCommandLoop();
    
    return TCL_OK;
}

static void
pltCommandLoop() 
{
    Tcl_DString command;	/* buffer incomplete commands being read
				 * from stdin. */
    int gotPartial, tty, code;
    char *buffer, *cmd;

#ifdef WIN32
    void xdMessageBox (char *string);
#endif

    tty = isatty(0);
    /*
     * Process commands from stdin until there's an end-of-file.
     */
    gotPartial = 0;
    Tcl_DStringInit(&command);
    while (1) {
	if ( gotPartial ) {
	    buffer = getCommand( "_> " );
	} else {
	    buffer = getCommand("xdgraph> ");
	}

	cmd = Tcl_DStringAppend(&command, buffer, -1);
	if ((buffer[0] != 0) && !Tcl_CommandComplete(cmd)) {
	    gotPartial = 1;
	    continue;
	}

	gotPartial = 0;
	code = Tcl_RecordAndEval(interp, cmd, 0);
/*	code = Tcl_Eval(interp, cmd);*/
	Tcl_DStringFree(&command);
	if (code != TCL_OK) {

#ifdef WIN32
            xdMessageBox(interp->result);
#else
	    fprintf(stderr, "%s\n", interp->result);
#endif

	} else if (tty && (*interp->result != 0)) {
            printf("%s\n", interp->result);
	}
    }
}

int
pltDriverCmd(ClientData data,
             Tcl_Interp *interp,            /* Current interpreter. */
             int argc,                      /* Number of arguments. */
             char **argv)                   /* Argument strings.    */
{
    if (driver.type == NODRIVER) {
        Tcl_SetResult(interp, "no driver selected", TCL_STATIC);
        return TCL_ERROR;
    }
    if (argc == 1) {
        switch (driver.type) {
#ifdef CC_OUTPUT
        case CC: {
            Tcl_SetResult(interp, "cc", TCL_STATIC);
        }; break;
#endif
#ifdef TK_OUTPUT
        case TK: {
            Tcl_SetResult(interp, "tk", TCL_STATIC);
        }; break;
#endif
#ifdef GT_OUTPUT
        case GT: {
            Tcl_SetResult(interp, "gt", TCL_STATIC);
        }; break;
#endif
        }
    }
    if (strEQ(argv[1], "features", 1)) {
        switch (driver.type) {
#ifdef CC_OUTPUT
        case CC: {
            Tcl_SetResult(interp, "contour relief path molecule", TCL_STATIC);
        }; break;
#endif
#ifdef TK_OUTPUT
        case TK: {
            Tcl_SetResult(interp, "xydiagram contour relief path molecule", TCL_STATIC);
        }; break;
#endif
#ifdef GT_OUTPUT
        case GT: {
            Tcl_SetResult(interp, "xydiagram contour relief path heightfield isosurface molecule", TCL_STATIC);
        }; break;
#endif
        }
    } else if (strEQ(NULL, "configure", 1)) {
        switch (driver.type) {
#ifdef GT_OUTPUT
        case GT: {
            return gtOptionsCmd(data, interp, argc-2, argv+2);
        }; break;
#endif
        }
    }
    return TCL_OK;
}

int
pltFlush(ViewP view)
{
        switch (driver.type) {
#ifdef CC_OUTPUT
    case CC: {
        ccmove((DrvgtP)view->drdat, 0., 0.)
    }; break;
#endif
#ifdef TK_OUTPUT
    case TK: {
        tkFlush((DrvtkP)view->drdat, NULL, 0);
    }; break;
#endif
#ifdef GT_OUTPUT
    case GT: {
    }; break;
#endif
    }
    return TCL_OK;
}


MarkDataT *
pltGetMarker(int m)
{
    if (m < 0 || m >= NUM_MARKER) {
        return NULL;
    } else {
        return Marker[m];
    }
}
