#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/pltObjects.c,v 2.16 1999/06/03 12:45:30 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: pltObjects.c,v $
 * Revision 2.16  1999/06/03 12:45:30  richter
 * pltScale um skalare Werte im view skalieren zu koennen
 *
 * Revision 2.15  1996/11/06 13:28:15  richter
 * (pltMessage): fixed core dump when a large string was received
 *
 * Revision 2.14  1996/11/06  10:59:25  richter
 * (pltSegment): linesets where not drawn correctly
 *
 * Revision 2.13  1996/09/14  13:20:18  richter
 * drawline wasn't always returning a status code
 *
 * Revision 2.12  1996/09/06  17:51:52  richter
 * moveit and drawit return an error status
 *
 * Revision 2.11  1996/09/06  11:30:00  richter
 * not too much left: higher level code (like pltAtoms) moved to the
 * corresponding command implementation (molCommand.c), lower level
 * code moved into the device driver
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.4  1996/07/25  06:05:17  richter
 * an unclean style was filled for atoms -- call styClean before styFillFG
 * new arguments to gtHeightField
 *
 * Revision 2.9.1.3  1996/07/03  10:13:08  richter
 * fixed zlimit test for transforming drivers
 * draw atom labels via <driver>Text, not <driver>Atom
 *
 * Revision 2.9.1.2  1996/05/20  11:17:54  richter
 * new transformation handling
 * pltText and plt?Axis routines
 *
 * Revision 2.9.1.1  1996/04/29  12:18:53  richter
 * last ci before transformation reorg.
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:09  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: pltDispatch
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration			*/
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "drvProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "strProto.h"
#include <stdarg.h>

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int
drawline(ViewP view,
         int (*moveit)(void *,double,double),
         int (*drawit)(void *,double,double),
	 double x1,
	 double y1,
	 int mode);
/*
 * variables with internal linkage
 */
static int     PatternLength = 0;
static double *PatternValues = NULL;

/*
 * variables with external linkage
 */
extern TheDriver driver;

/* VARARGS0 */
int
pltMessage(char *level,
	   char *fmt,
	   ...)
{
    va_list ap;
    int code;

    va_start(ap, fmt);
    if (strlen(fmt) > PRINT_BUFFER_SIZE) {
        return TCL_OK;
    }
    
    switch (driver.type) {
#ifdef TK_OUTPUT
    case GT:
    case TK: {
        vsprintf(buf, fmt, ap);
        code = Tcl_VarEval(interp, "logMessage ", level, " {", buf, "}", NIL);
        if (code == TCL_ERROR) {
            Tcl_AddErrorInfo(interp, "(pltLogInfo)\n");
        }
    }; break;
#endif
    default: {
	if (strEQ(level,"err",3)) {
	    vfprintf(stderr, fmt, ap);
	} else if (!strEQ(NULL,"cmd",3)) {
	    vfprintf(stdout, fmt, ap);
	}
    }; break;
    }
    va_end(ap);
    return code;
}

void
pltTransform(ViewP view,
             c3dT *ci,
             c3dT *co)
{
    int i, k;
    double coo[3], co2[3];
    
    coo[0] = ci->x;
    coo[1] = ci->y;
    coo[2] = ci->z;

    if (view->opt.doorig) {
        for (i=0; i<3; ++i) coo[i] -= view->opt.orig[i];
    }

    if (view->opt.dorot) {
        for (i=0; i<3; ++i) {
            co2[i] = 0.;
            for (k=0; k<3; ++k) co2[i] += view->opt.rotm[i][k] * coo[k];
        }
        for (i=0; i<3; ++i) coo[i] = co2[i];
    }

    if (view->opt.dotrans) {
        for (i=0; i<3; ++i) coo[i] += view->opt.trans[i];
    }

    if (view->opt.doscale) {
        for (i=0; i<3; ++i) coo[i] *= view->opt.scale;
    }

    co->x = coo[0];
    co->y = coo[1];
    co->z = coo[2];

    return;
}

double
pltScale(ViewP view, double v)
{
    double s = v;
    if (view->opt.doscale) {
        s *= view->opt.scale;
    }
    return s;
}

int
pltSegment(ViewP view,
           SegmentP segment,
           int (*moveit)(void *,double,double),
           int (*drawit)(void *,double,double))
{
    SegmentP s;
    int n, mode, code;
    c3dT crd, cra, crb;

    s = segment;
    if ( s->type == INVALIDSEG || s->np <= 0 ) return TCL_OK;
    if (s->type == LINESET) {
        for ( n=0; n < s->np; n++ ) {
            cra.x = s->p.line.a[n].x;
            cra.y = s->p.line.a[n].y;
            cra.z = s->p.line.a[n].z;
            crb.x = s->p.line.b[n].x;
            crb.y = s->p.line.b[n].y;
            crb.z = s->p.line.b[n].z;
            if (!driver.transforms) {
                pltTransform(view, &cra, &cra);
                pltTransform(view, &crb, &crb);
            }
            if ((code=drawline(view, moveit, drawit, cra.x, cra.y, 0))
                != TCL_OK) return code;
            if ((code=drawline(view, moveit, drawit, crb.x, crb.y, 1))
                != TCL_OK) return code;
        }
    } else {
        mode = (s->type == DRAWSEG);
        for ( n=0; n < s->np; n++ ) {
            crd.x = s->p.point[n].x;
            crd.y = s->p.point[n].y;
            crd.z = s->p.point[n].z;
            if (!driver.transforms) {
                pltTransform(view, &crd, &crd);
            }
            if ((code=drawline(view, moveit, drawit, crd.x, crd.y, mode))
                != TCL_OK) return code; 
           mode = 1;
        }
    }
    return TCL_OK;
}

static int
drawline(ViewP view,
         int (*moveit)(void *,double,double),
         int (*drawit)(void *,double,double),
	 double x1,
	 double y1,
	 int mode)		/* 0=move, 1=draw */
{
    int code;
    static double x0 = 0.;
    static double y0 = 0.;
    static double lastu = 0.;
    static int lastp = 0;

    double x, y, dx, dy, dabs, use, dleft;

    dx = x1 - x0;
    dy = y1 - y0;
    if ( mode==0 ) {
	if ((code=moveit(view->drdat, x1, y1)) != TCL_OK) return code;
	lastu = 0.;
	lastp = 0;
	x0 = x1;
	y0 = y1;
	return TCL_OK;
    } else if ( PatternLength == 0 ) {
	drawit(view->drdat, x1, y1);
	lastu = 0.;
	lastp = 0;
	x0 = x1;
	y0 = y1;
	return TCL_OK;
    }

    dabs = sqrt( dx*dx + dy*dy );
    if ( dabs < 1.e-6 ) {
	return TCL_OK;
    }
    dleft = dabs;
    x = x0;
    y = y0;
    while ( dleft > (use = fabs(PatternValues[lastp]) - lastu) ) {
	x = x + use * dx / dabs;
	y = y + use * dy / dabs;
	if ( PatternValues[lastp] < 0. ) {
	    if ((code=moveit(view->drdat, x, y)) != TCL_OK) return code;
	} else {
	    drawit(view->drdat, x, y);
	}
	lastu = 0.;
	lastp = (lastp+1) % PatternLength;
	dleft -= use;
    }
    use = dleft;
    x = x + use * dx / dabs;
    y = y + use * dy / dabs;
    if ( PatternValues[lastp] < 0. ) {
	if ((code=moveit(view->drdat, x, y)) != TCL_OK) return code;
    } else {
	drawit(view->drdat, x, y);
    }
    lastu += use;
    
    if ( fabs(PatternValues[lastp]) - lastu <= 1.e-6 ) {
	lastp = (lastp+1) % PatternLength;
	lastu =0.;
    }
    x0 = x;
    y0 = y;
    return TCL_OK;
}

int
pltSetLinestyle(LineP lPtr)
{
    int i;

    if (PatternValues) {
        ckfree((char *)PatternValues);
    }
    PatternLength = lPtr->plength;
    if (PatternLength > 0) {
        PatternValues = (double *) ckalloc(lPtr->plength*sizeof(double));
        for (i=0; i<lPtr->plength; ++i) {
            PatternValues[i] = lPtr->pattern[i];
        }
    } else {
        PatternValues = (double *) 0;
    }
    return TCL_OK;
}

int
pltRemove(ViewP view,
          int tag)
{
    switch (driver.type) {
#ifdef CC_OUTPUT
        case CC: {
        }; break;
#endif
#ifdef TK_OUTPUT
        case TK: {
            return tkRemove(view, tag);
        }; break;
#endif
#ifdef GT_OUTPUT
        case GT: {
        }; break;
#endif
    }
    return TCL_OK;
}
