#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/relCommand.c,v 2.14 1997/03/29 18:10:47 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 *======================================================================
 *
 * Module: relCommand
 *
 * Abstract: Defines the 'relief' command and creates relief object
 * commands.  Called from the tcl interpreter.
 *
 * $Log: relCommand.c,v $
 * Revision 2.14  1997/03/29 18:10:47  richter
 * preset style
 *
 * Revision 2.13  1996/11/03  16:29:29  richter
 * styGetConfig expects a style group mask
 *
 * Revision 2.12  1996/09/11  11:24:25  richter
 * :relief configure did not work correctly
 *
 * Revision 2.11  1996/09/06  11:51:12  richter
 * many changes to use the new display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.10.1.1  1996/07/25  08:51:00  richter
 * *** empty log message ***
 *
 * Revision 2.10  1996/07/25  08:50:44  richter
 * *** empty log message ***
 *
 * Revision 2.10  1996/07/25  08:48:31  richter
 * *** empty log message ***
 *
 * Revision 2.9.1.3  1996/07/25  06:15:06  richter
 * use Tcl_SetVar2 to set active(relief)
 * handle 'relief <relief>' correctly
 *
 * Revision 2.9.1.2  1996/07/03  10:56:56  richter
 * refined view scaling
 * configure output
 * -eye and -size take lists as values
 *
 * Revision 2.9.1.1  1996/05/20  11:21:54  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:10  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "vieProto.h"
#include "cmdProto.h"
#include "pltProto.h"
#include "relProto.h"
#include "setProto.h"
#include "styProto.h"
#include "strProto.h"
#include "vieProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureRelief(Tcl_Interp *interp, ReliefP rPtr,
			   int argc,char **argv, ViewOptT *vOpt, int recalc);
static int showcfgRelief(Tcl_Interp *interp, ReliefP rPtr);
static int plotRelief(Tcl_Interp *interp, ViewP view, ReliefP rPtr);
static int clearRelief(Tcl_Interp *interp, ViewP view, ReliefP rPtr);

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
ReliefCmd(ClientData dummy,		/* Not used			*/
	  Tcl_Interp *interp,		/* Current interpreter		*/
	  int argc,			/* Number of arguments		*/
	  char **argv)			/* Argument strings		*/
{
    ReliefP  rPtr;
    DatasetP dPtr;
    GridP gPtr;
    ViewOptT vOpt;
    char *objName;
    ViewE vieType = VIEW_RELIEF;

    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
        return showcfgRelief(interp, NULL);
    }
    
    if (cmdAnalyzeName(argv[1], &objName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }

    if (dPtr->type == DSET_GRID) {
        gPtr = &dPtr->data.dg->grid;
    } else if (dPtr->type != DSET_SLICE) {
        gPtr = &dPtr->data.ds->grid;
    } else {
	Tcl_AppendResult(interp,
                         "relief is only valid for GRID or SLICE datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }
        
    rPtr = relAlloc(argv[1]);
    if (!rPtr) {
	Tcl_AppendResult(interp, "couldn't create relief", (char *) NULL);
	return TCL_ERROR;
    }
    rPtr->object = strDupl(objName);
    rPtr->vertac = ((double) (gPtr->nx + gPtr->ny)) /
	fabs(4.*(gPtr->vmax - gPtr->vmin));
    rPtr->hcut = gPtr->vmax;
    rPtr->lcut = gPtr->vmin;
    rPtr->nx = gPtr->nx;
    rPtr->ny = gPtr->ny;
    styFillFG(interp, &rPtr->style, "black");
    styFillLinewidth(interp, &rPtr->style, 0.5);
    styFillPattern(interp, &rPtr->style, "solid");

    vOpt.doscale = 1;
    vOpt.doorig  = 0;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = -1.;
    vOpt.width = -1.;
    vOpt.height = -1.;
    if (configureRelief(interp, rPtr, argc-2, argv+2, &vOpt, TRUE) != TCL_OK) {
	goto Error;
    }

    vOpt.vrange[0] = dPtr->xsize;
    vOpt.vrange[1] = dPtr->ysize;
    if (cmdSetView(interp, &rPtr->view, 1, &vieType, objName,
                   argv[1], "relief", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }

    Tcl_CreateCommand(interp, argv[1], ReliefObjectCmd,
		      (ClientData)rPtr, (Tcl_CmdDeleteProc *)NULL );
    Tcl_SetVar2(interp, "active", "relief", argv[1], TCL_GLOBAL_ONLY);
    return TCL_OK;

Error:
    relFree(&rPtr);
    return TCL_ERROR;
}

int
ReliefObjectCmd(ClientData data,		/* Relief		*/
		 Tcl_Interp *interp,		/* Current interpreter. */
		 int argc,			/* Number of arguments. */
		 char **argv)			/* Argument strings.	*/
{
    ReliefP  rPtr;
    ViewP view;
    ViewOptT vOpt;

    rPtr = (ReliefP) data;
    view = viewFindByName(rPtr->view, NULL);

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"",
			 (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1],"configure",2)) {
        if (argc <= 2) {
            return showcfgRelief(interp, rPtr);
        } else {
            return configureRelief(interp, rPtr, argc-2, argv+2, &vOpt, FALSE);
        }
    } else if (strEQ(NULL,"clear",2)) {
        clearRelief(interp, view, rPtr);
    } else if (strEQ(NULL,"info",1)) {
	return relInfo(interp, rPtr);
    } else if (strEQ(NULL,"plot",1)) {
        clearRelief(interp, view, rPtr);
        return plotRelief(interp, view, rPtr);
    } else if (strEQ(NULL,"remove",1)) {
        clearRelief(interp, view, rPtr);
        viewUnregisterObject(view, rPtr->name);
	Tcl_DeleteCommand(interp, rPtr->name);
	relFree(&rPtr);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, info, clear, plot or remove",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;

}

static int
configureRelief(Tcl_Interp *interp,
		ReliefP rPtr,
		int argc,
		char **argv,
                ViewOptT *vOpt,
		int recalc)
{
    int i;
    int has2styles;
    double value;
    int elemc;
    char **elemv;
    GridP gPtr;
    DatasetP dPtr;
    
    dPtr = setFindByName(rPtr->object);
    if (dPtr->type == DSET_GRID) {
        gPtr = &dPtr->data.dg->grid;
    } else {
        gPtr = &dPtr->data.ds->grid;
    }

    viewParseArgs(NULL, &argc, argv, vOpt);
    
    styParseArgs(&argc, argv, &rPtr->style, NULL, &has2styles);

    for (i=0; i<argc; ++i) {
	if (strEQ(argv[i],"-cutoff",2)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-cutoff\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
            if ( Tcl_SplitList(interp, argv[++i], &elemc, &elemv) != TCL_OK ) {
		return TCL_ERROR;
	    }
            if (elemc == 1) {
                if ( Tcl_GetDouble(interp, elemv[0], &value) != TCL_OK ) {
                    return TCL_ERROR;
                }
                rPtr->hcut = value;
                rPtr->lcut = -value;
            } else {
                if ( Tcl_GetDouble(interp, elemv[0], &rPtr->hcut) != TCL_OK ) {
                    return TCL_ERROR;
                }
                if ( Tcl_GetDouble(interp, elemv[1], &rPtr->lcut) != TCL_OK ) {
                    return TCL_ERROR;
                }
            }
            recalc = TRUE;
	} else if (strEQ(NULL,"-eye",2)) {
	    if (++i >= argc) {
		Tcl_AppendResult(interp, "value for \"-eye\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
		return TCL_ERROR;
	    }
            if (elemc != 3) {
                Tcl_AppendResult(interp, "need list of 3 values for \"-eye\"",
				 (char *) NULL);
		return TCL_ERROR;
            }
            if ( Tcl_GetDouble(interp, elemv[0], &rPtr->eye.x) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, elemv[1], &rPtr->eye.y) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, elemv[2], &rPtr->eye.z) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    recalc = TRUE;
	} else if (strEQ(NULL,"-hcut",3)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-hcut\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[++i], &value) != TCL_OK  ) {
		return TCL_ERROR;
	    }
	    rPtr->hcut = value;
	    recalc = TRUE;
	} else if (strEQ(NULL,"-hsize",3)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-hsize\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[++i], &rPtr->nx) != TCL_OK  ) {
		return TCL_ERROR;
	    }
	    recalc = TRUE;
	} else if (strEQ(NULL,"-lcut",2)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-lcut\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[++i], &value) != TCL_OK  ) {
		return TCL_ERROR;
	    }
	    rPtr->lcut = value;
	    recalc = TRUE;
	} else if (strEQ(NULL,"-size",3)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-size\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
            if ( Tcl_SplitList(interp, argv[++i], &elemc, &elemv) != TCL_OK ) {
		return TCL_ERROR;
	    }
            if (elemc != 2) {
                Tcl_AppendResult(interp, "need list of 2 values for \"-size\"",
				 (char *) NULL);
		return TCL_ERROR;
            }
            if ( Tcl_GetInt(interp, elemv[0], &rPtr->nx) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, elemv[1], &rPtr->ny) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    recalc = TRUE;
	} else if (strEQ(NULL,"-vertac",3)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-vertac\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[++i], &rPtr->vertac) != TCL_OK  ) {
		return TCL_ERROR;
	    }
	    recalc = TRUE;
	} else if (strEQ(NULL,"-vsize",3)) {
	    if ( i+1 >= argc ) {
		Tcl_AppendResult(interp, "value for \"-vsize\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[++i], &rPtr->ny) != TCL_OK  ) {
		return TCL_ERROR;
	    }
	    recalc = TRUE;	    
        } else if (strEQ(NULL,"-view",3)) {
            i++;
            if (i >= argc) {
                Tcl_AppendResult(interp, "value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            rPtr->view = strDupl(argv[i]);
	} else {
	    Tcl_AppendResult(interp, "unknown option \"",
			     argv[i], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (recalc) {
	relEngine(gPtr, rPtr);
    }
    return TCL_OK;
		
}

static int
showcfgRelief(Tcl_Interp *interp,
              ReliefP rPtr)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (rPtr) {
        Tcl_DStringAppendElement(&cfg, rPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "cutoff");
    Tcl_DStringAppendElement(&cfg, "float:2::high,low");
    Tcl_DStringAppendElement(&cfg, "Cutoff Values");
    Tcl_DStringAppendElement(&cfg, "");
    if (rPtr) {
        Tcl_DStringStartSublist(&cfg);
        sprintf(buf, "%g", rPtr->hcut);
        Tcl_DStringAppendElement(&cfg, buf);
        sprintf(buf, "%g", rPtr->lcut);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "eye");
    Tcl_DStringAppendElement(&cfg, "float:3::x,y,z");
    Tcl_DStringAppendElement(&cfg, "Eye Point");
    Tcl_DStringAppendElement(&cfg, "1. 1. 0.6");
    if (rPtr) {
        Tcl_DStringStartSublist(&cfg);
        sprintf(buf, "%g", rPtr->eye.x);
        Tcl_DStringAppendElement(&cfg, buf);
        sprintf(buf, "%g", rPtr->eye.y);
        Tcl_DStringAppendElement(&cfg, buf);
        sprintf(buf, "%g", rPtr->eye.z);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "size");
    Tcl_DStringAppendElement(&cfg, "float:2::nx,ny");
    Tcl_DStringAppendElement(&cfg, "Size");
    Tcl_DStringAppendElement(&cfg, "");
    if (rPtr) {
        Tcl_DStringStartSublist(&cfg);
        sprintf(buf, "%d", rPtr->nx);
        Tcl_DStringAppendElement(&cfg, buf);
        sprintf(buf, "%d", rPtr->ny);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "vertac");
    Tcl_DStringAppendElement(&cfg, "float");
/* Cljf    Tcl_DStringAppendElement(&cfg, "Vertical Acceleration"); */ 
    Tcl_DStringAppendElement(&cfg, "Vertical Accentuation"); 
    Tcl_DStringAppendElement(&cfg, "");
    if (rPtr) {
        sprintf(buf, "%g", rPtr->vertac);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    if (rPtr) {
        ViewP view = viewFindByName(rPtr->view, NULL);
        styGetConfig(&cfg, &rPtr->style, STY_FG_ALL|STY_LINE_ALL, 0, 0, NULL);
        viewGetConfig(&cfg, &view->opt, 0);
    } else {
        styGetConfig(&cfg, NULL, STY_FG_ALL|STY_LINE_ALL, 0, 0, NULL);
        viewGetConfig(&cfg, NULL, 0);
    }

    if (rPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_flags");
        sprintf(buf, "%d", rPtr->flags);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
        
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_object");
        Tcl_DStringAppendElement(&cfg, rPtr->object);
        Tcl_DStringEndSublist(&cfg);
    }
    
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
plotRelief(Tcl_Interp *interp,
           ViewP view,
           ReliefP rPtr)
{
    if (!view) return TCL_ERROR;

#ifdef DEBUG
    pltMessage("inf", "relief ");
#endif
    if (!rPtr->tag) {
        rPtr->tag = dlAlloc(view, DL_SEGMENT, DL_MAIN_PLANE);
        dlMkSegment(rPtr->tag, rPtr->seg, &rPtr->style);
    }
    pltUpdateView(view);
    return TCL_OK;
}

static int
clearRelief(Tcl_Interp *interp,
            ViewP view,
            ReliefP rPtr)
{
    if (!view) return TCL_ERROR;
    if (rPtr->tag) {
        dlRemove(view, rPtr->tag);
        rPtr->tag = 0;
    }
    return TCL_OK;
}
