#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/segManager.c,v 2.10 1996/07/25 09:14:52 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: segManager.c,v $
 * Revision 2.10  1996/07/25 09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.1  1996/05/19  16:00:59  richter
 * enhanced segments
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:12  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "segProto.h"
#include "pltProto.h"

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static int segmentAlloc = 100;

SegmentP
segAlloc(int size,
         int what)
{
    SegmentP seg;
    int sa;

    if ( size > 0 ) {
	sa = size;
    } else {
        sa = segmentAlloc;
    }
    seg = (SegmentP) ckalloc( sizeof(SegmentT) );
    seg->next  = NULL;
    seg->alloc = sa;
    seg->np    = 0;
    switch (what) {
    case LINESET: {
        seg->p.line.a = (c3dT *) ckalloc( sa * sizeof(c3dT) );
        seg->p.line.b = (c3dT *) ckalloc( sa * sizeof(c3dT) );
        seg->type = LINESET;
    }; break;
    case POINTSET: {
        seg->p.point = (c3dT *) ckalloc( sa * sizeof(c3dT) );
        seg->type = POINTSET;
    }; break;
    default: {
        seg->p.point = (c3dT *) ckalloc( sa * sizeof(c3dT) );
        seg->type = INVALIDSEG;
    }; break;
    }
    return(seg);
}

void
segFree( SegmentP *seg )
{
    SegmentP s;

    s = *seg;
    if ( ! s ) {
	return;
    }
    if (s->type == LINESET) {
        ckfree( (char *)s->p.line.a );
        ckfree( (char *)s->p.line.b );
    } else if ( s->p.point ) {
	ckfree( (char *)s->p.point );
    }
    if ( s->next ) {
	segFree( &s->next );
    }
    ckfree( (char *)s );
    *seg = (SegmentP) NULL;
    return;
}

void
segAddSegment(SegmentP *seg,
              double x,
              double y,
              double z)
{
    SegmentP s, oldseg;

    s = *seg;
    if ( s->np >= s->alloc ) {
	oldseg = *seg;
	if ( s->next ) {
	    s = *seg = s->next;
	} else {
	    s->next = (SegmentP) ckalloc( sizeof(SegmentT) );
	    if ( !s->next ) {
		pltMessage("err", "Memory allocation error - abort\n");
		exit(1);
	    }
	    s = *seg = s->next;
	    s->p.point = (c3dT *) ckalloc( segmentAlloc * sizeof(c3dT) );
	    s->next  = NULL;
	    s->alloc = segmentAlloc;
	}
	s->type  = DRAWSEG;
	--oldseg->np;
	s->p.point[0] = oldseg->p.point[oldseg->np];
	s->np    = 1;
    } else if ( s->type == INVALIDSEG ) {
	s->type  = MOVESEG;
    }
    s->p.point[s->np].x = x;
    s->p.point[s->np].y = y;
    s->p.point[s->np].z = z;
    s->np++;
    return;
}

void
segAddLine(SegmentP *seg,
           double x1, double y1, double z1,
           double x2, double y2, double z2)
{
    SegmentP s;

    s = *seg;
    if ( s->np >= s->alloc ) {
        s->next = segAlloc(s->alloc, LINESET);
        if ( !s->next ) {
            pltMessage("err", "Memory allocation error - abort\n");
            exit(1);
        }
        s = *seg = s->next;
    }
    s->p.line.a[s->np].x = x1;
    s->p.line.a[s->np].y = y1;
    s->p.line.a[s->np].z = z1;
    s->p.line.b[s->np].x = x2;
    s->p.line.b[s->np].y = y2;
    s->p.line.b[s->np].z = z2;
    s->np++;
    return;
}

void
segAddPoint(SegmentP *seg,
           double x, double y, double z)
{
    SegmentP s;

    s = *seg;
    if ( s->np >= s->alloc ) {
        s->next = segAlloc(s->alloc, POINTSET);
        if ( !s->next ) {
            pltMessage("err", "Memory allocation error - abort\n");
            exit(1);
        }
        s = *seg = s->next;
    }
    s->p.point[s->np].x = x;
    s->p.point[s->np].y = y;
    s->p.point[s->np].z = z;
    s->np++;
    return;
}

void
segSwitchSegment(SegmentP *seg)
{
    SegmentP s;
    c3dT temp;

    s = *seg;
    if (s->np < 2) return;

    temp = s->p.point[s->np-1];
    s->p.point[s->np-1] = s->p.point[s->np-2];
    s->p.point[s->np-2] = temp;    
    return;
}

void
segReplaceSegment(SegmentP *seg,
	    double x,
	    double y,
	    double z)
{
    SegmentP s;

    s = *seg;
    if (s->np < 2) return;
    s->p.point[s->np-1].x = x;
    s->p.point[s->np-1].y = y;
    s->p.point[s->np-1].z = z;
    return;
}

void
segFinishSegment( SegmentP *seg )
{
    SegmentP s;

    s = *seg;
    if (s->type == DRAWSEG || s->np > 1) {
	s->next = (SegmentP) ckalloc( sizeof(SegmentT) );
	if ( !s->next ) {
	    pltMessage("err", "Memory allocation error - abort\n");
	    exit(1);
	}
	s = *seg = s->next;
	s->p.point = (c3dT *) ckalloc( segmentAlloc * sizeof(c3dT) );
	s->next  = NULL;
	s->alloc = segmentAlloc;
    }
    s->np   = 0;
    s->type = INVALIDSEG;
    return;
}

void
segDump( SegmentP seg )
{
    int i;
    SegmentP s;

    for ( s=seg; s; s=s->next) {
	if ( s->type == INVALIDSEG || s->np <= 0 ) {
	    printf( "INVALIDSEG %d %d %d\n", s->type, s->alloc, s->np );
	    continue;
	}
	if (s->type == DRAWSEG) {
	    printf( "DRAWSEG %d %d %d\n", s->type, s->alloc, s->np );
	} else {
	    printf( "MOVESEG %d %d %d\n", s->type, s->alloc, s->np );
	}
	for ( i=0; i < s->np; i++ ) {
	    printf("\t%g %g %g\n",
		   s->p.point[i].x, s->p.point[i].y, s->p.point[i].z);
	}
    }
    return;
}
