#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/styApplyManager.c,v 2.12 1996/11/03 16:38:28 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 *======================================================================
 *
 * Module: styManager
 *
 * $Log: styApplyManager.c,v $
 * Revision 2.12  1996/11/03 16:38:28  richter
 * added styApplyPolygon
 *
 * Revision 2.11  1996/09/06  11:56:35  richter
 * - no static style anymore: fill an (empty) style structure according to
 *   a current style, a new style and a mask specifying which values to use
 * - new routine styApplyInterpol (for colour interpolation)
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9  1996/07/03  14:32:57  richter
 * *** empty log message ***
 *
 * Revision 2.9.1.1  1996/05/19  16:03:43  richter
 * new routine styParseArgs
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:16  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "strProto.h"
#include "styProto.h"
#include "pltProto.h"
#ifdef VMS
#include "../../libxd/src/strcasecmp.h"
#endif

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * initialize variables with external linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * functions with external and internal linkage
 *----------------------------------------------------------------------
 */


int
styApply(StyleP out,
         int *out_mask,
         StyleP current,
         StyleP in,
         int in_mask)
{
    styApplyForeground(out, out_mask, current,
                       in_mask&STY_FG_NAME, in->fg.name,
                       in_mask&STY_FG_MODEL, in->fg.model,
                       in_mask&STY_FG_VALUE, in->fg.val);
    styApplyLine(out, out_mask, current,
                in_mask&STY_LINE_WIDTH, in->line.width,
                in_mask&STY_LINE_PATTERN, in->line.plength, in->line.pattern);
    styApplyFont(out, out_mask, current,
                 in_mask&STY_FONT_SIZE, in->font.size,
                 in_mask&STY_FONT_NAME, in->font.name,
                 in_mask&STY_FONT_ANCHOR, in->font.anchor);
    styApplyPolygon(out, out_mask, current,
                    in_mask&STY_POLYGON_MODE, in->pg.mode);
    return TCL_OK;
}

int
styApplyForeground(StyleP out,
                   int *out_mask,
                   StyleP cur,
                   int fgnam,
                   char *name,
                   int fgmod,
                   ColorE model,
                   int fgval,
                   float val[3])
{
    int i;
    
    if (fgnam) {
        if (name) {
            if (cur->fg.name == NULL || strcmp(name, cur->fg.name) !=0 ) {
                cur->fg.name = strDuplOver(cur->fg.name, name);
                out->fg.name = strDupl(name);
                *out_mask |= STY_FG_NAME; 
            }
	} else {
	    if (cur->fg.name) {
		ckfree(cur->fg.name);
		cur->fg.name = NULL;
                *out_mask |= STY_FG_NAME;
	    }
	}
    }

    if (fgmod && model != cur->fg.model) {
        out->fg.model = cur->fg.model = model;
        *out_mask |= STY_FG_MODEL;
    } else {
        out->fg.model = cur->fg.model;
    }

    if (fgval && memcmp(val, cur->fg.val, 3*sizeof(float))!=0) {
        for (i=0; i<3; ++i) {
            out->fg.val[i] = cur->fg.val[i] =  val[i];
        }
        *out_mask |= STY_FG_VALUE;
    }

    return TCL_OK;
}

int
styApplyLine(StyleP out,
             int *out_mask,
             StyleP cur,
             int lnwid,
             double width,
             int lnpat,
             int plength,
             double *pattern)
{
    int i;

    if (lnwid && width != cur->line.width) {
        out->line.width = cur->line.width = width;
        *out_mask |= STY_LINE_WIDTH;
    }

    if (lnpat) {
	lnpat = FALSE;
	if (plength != cur->line.plength) {
            lnpat = TRUE;
	} else {
	    if (pattern) {
		if (cur->line.pattern) {
		    lnpat = memcmp(pattern, cur->line.pattern,
                                   cur->line.plength*sizeof(double));
		} else {
		    lnpat = TRUE;
		}
	    } else if (cur->line.pattern) {
		lnpat = TRUE;
	    }
	}
	if (lnpat) {
	    if (cur->line.pattern) {
		ckfree((char *)cur->line.pattern);
	    }
	    out->line.plength = cur->line.plength = plength;
	    if (plength > 0) {
		cur->line.pattern = (double *) ckalloc(plength*sizeof(double));
                out->line.pattern = (double *) ckalloc(plength*sizeof(double));
		for (i=0; i<plength; ++i) {
		    out->line.pattern[i] = cur->line.pattern[i] = pattern[i];
		}
	    } else {
		cur->line.pattern = (double *) 0;
	    }
            *out_mask |= STY_LINE_PATTERN;
	}
    }

    return TCL_OK;
}

int
styApplyPolygon(StyleP out,
                int *out_mask,
                StyleP cur,
                int pgmode,
                PolygonE mode)
{
    int i;

    if (pgmode && mode != cur->pg.mode) {
        out->pg.mode = cur->pg.mode = mode;
        *out_mask |= STY_POLYGON_MODE;
    }
    return TCL_OK;
}

int
styApplyFont(StyleP out,
             int *out_mask,
             StyleP cur,
             int fosiz,
             double size,
             int fonam,
             char *name,
             int foanc,
             char *anchor)
{
    if (fosiz && size != cur->font.size) {
        out->font.size = cur->font.size = size;
        *out_mask |= STY_FONT_SIZE;
    }

    if (fonam) {
	if (name) {
	    if (cur->font.name == NULL ||
		strcmp(name, cur->font.name)!=0) {
		cur->font.name =strDuplOver(cur->font.name, name);
		out->font.name = strDupl(name);
                out->font.size = cur->font.size;
                *out_mask |= STY_FONT_NAME;
	    }
	} else {
	    if (cur->font.name) {
		ckfree(cur->font.name);
		cur->font.name = NULL;
                *out_mask |= STY_FONT_NAME;
	    }
	}
    }

    if (foanc) {
	if (anchor) {
	    if (!cur->font.anchor ||
		strcmp(anchor, cur->font.anchor)!=0) {
		cur->font.anchor = strDuplOver(cur->font.anchor, anchor);
                out->font.anchor = strDupl(anchor);
                *out_mask |= STY_FONT_ANCHOR;
	    }
	} else {
	    if (cur->font.anchor) {
		ckfree(cur->font.anchor);
		cur->font.anchor = NULL;
                *out_mask |= STY_FONT_ANCHOR;
	    }
	}
    }

    return TCL_OK;
}
