#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/styFillManager.c,v 2.14 1997/03/29 18:11:25 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 *======================================================================
 *
 * Module: styManager
 *
 * $Log: styFillManager.c,v $
 * Revision 2.14  1997/03/29 18:11:25  richter
 * styFillLinewidth returns an error code
 *
 * Revision 2.13  1996/11/06  11:00:47  richter
 * (styFillAnchor): new routine
 *
 * Revision 2.12  1996/11/03  16:43:00  richter
 * added styFillPolygon()
 *
 * Revision 2.11  1996/09/06  11:58:16  richter
 * new routine styFillInterpol (formerly styApplyInterpol)
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9  1996/07/03  14:33:02  richter
 * *** empty log message ***
 *
 * Revision 2.9.1.1  1996/05/19  16:03:43  richter
 * new routine styParseArgs
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:16  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "strProto.h"
#include "styProto.h"
#include "pltProto.h"
#ifdef VMS
#include "../../libxd/src/strcasecmp.h"
#endif

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int styFillColorByName(Tcl_Interp *interp, ColorP col, char *name);
static int styFillColorByRGB(Tcl_Interp *interp, ColorP col, int iprec,
			    int rgb[3]);
static int styFillColorByValue(Tcl_Interp *interp, ColorP col, ColorE model,
			      char *val[3]);

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static char *colmodel[NCOLMODELS] = {"???", "rgb", "hsv", "yuv", "map"};


/*
 *----------------------------------------------------------------------
 * functions with external and internal linkage
 *----------------------------------------------------------------------
 */

int
styFillPolygon(Tcl_Interp *interp,
               StyleP sPtr,
               char *string)
{
    if (strcmp(string, "fill") == 0) {
        sPtr->pg.mode = PG_FILL;
    } else if (strcmp(string, "line") == 0) {
        sPtr->pg.mode = PG_LINE;
    } else if (strcmp(string, "point") == 0) {
        sPtr->pg.mode = PG_POINT;
    }
    return TCL_OK;
}

int
styFillAnchor(Tcl_Interp *interp,
              StyleP sPtr,
              char *anchor)
{
    sPtr->font.anchor = strDupl(anchor);
    return TCL_OK;
}

int
styFillLinewidth(Tcl_Interp *interp,
                 StyleP sPtr,
		 double value)
{
    sPtr->line.width = value;
    return TCL_OK;
}

int
styFillPattern(Tcl_Interp *interp,	/* TCL interpreter, used for error
					 * messages			    */
	      StyleP sPtr,		/* Style to be modified		    */
	      char *sstring)		/* Linestyle pattern in verbal form */
/*
 * Set the line pattern of a style.  The pattern is given in a verbal form:
 *
 * linestyle :== 'solid' | elements ...
 * elements  :== element [blanks]
 * element   :== 'dot' | 'dash' | 'long'
 * blanks    :== ' ' ...
 *
 * If no trailing blanks are specified, a default of one is used.
 * (Except for 'solid', of course).  Leading blanks are ignored.
 *
 * E.g. "solid"
 *	"dot"
 *	"dash dot  dot  dot"
 */
{
    static double fg[] = { 0.03, 0.20, 0.40 };  /* dot, dash, long */
    static double bg[] = { 0.07, 0.15, 0.25 };
    static double bi[] = { 0.03, 0.05, 0.05 };
    char *token, *ptk;
    int elem, first, last, curalloc, length, nbl;
    LineP lPtr;

    lPtr = &sPtr->line;
    lPtr->verbal = strDuplOver(lPtr->verbal, sstring); 
    /*
     * Setting the linetype to solid.
     * This is done by cleaning the pattern array and setting a
     * pattern length of zero.
     */
    if ( strcmp(sstring,"solid") == 0 ) {
	if ( lPtr->pattern ) {
	    ckfree((char *)lPtr->pattern);
	    lPtr->plength = 0;
	}
	return TCL_OK;
    }

    /*
     * Allocate a new pattern array.
     * TODO:  This should be done in a separate array, so that in case
     * of errors, the old values are retained.  Free the old space.
     */
    curalloc = 10;
    lPtr->pattern = (double *) ckalloc( curalloc * sizeof(double) );
    lPtr->plength = 0;
    first = 0;
    ptk   = sstring;
    length = strlen(sstring);
    last = 0;
    /*
     * Tokenize string and loop over tokens.
     */
    for( token=strtok(sstring," "); token; token=strtok(NULL," " ) ) {
	/*
	 * Take care for complicated patterns.
	 */
	if ( lPtr->plength+1 >= curalloc ) {
	    curalloc += 10;
	    lPtr->pattern = (double *) ckrealloc((char *)lPtr->pattern, curalloc);
	}
	/*
	 * Which type is used for this element?
	 */
	if ( strcmp(token,"dot") == 0 ) {
	    elem = 0;
	} else if ( strcmp(token,"dash") == 0 ) {
	    elem = 1;
	} else if ( strcmp(token,"long") == 0 ) {
	    elem = 2;
	} else {
	    Tcl_AppendResult(interp, "unrecognized string \"",
			     token, "\" in line style pattern", (char *) NULL);
	    return TCL_ERROR;
	}
	if (lPtr->plength <= 0) {
	    /*
	     * First token
	     */
	    first = elem;
	    /* lead  = token - sstring; TODO */
	} else {
	    /*
	     * Count number of blanks between this token and the
	     * previous one:
	     *
	     *     ptk      token
	     *     |        |
	     *     V        V
	     * "   token1   token2"
	     *  <+><-+--><+>
	     *   |   |    |
	     *   |   |    nbl
             *   |   strlen(ptk)
	     *   lead
	     */
	    nbl = (token - ptk) - strlen(ptk) - 1; /* + lead; TODO */
	    nbl = nbl > 0 ? nbl : 0;
	    /*
	     * Calculate blank ('background') length.
	     */
	    lPtr->pattern[lPtr->plength++] = - ((bg[elem]+bg[last])/2. +
					    nbl * (bi[elem]+bi[last])/2.);
	}
	/*
	 * Set line ('foreground') length.
	 */
	lPtr->pattern[lPtr->plength++] = fg[elem];
	last = elem;
	ptk  = token;
    }

    if ( lPtr->plength <= 0 ) {
	Tcl_AppendResult(interp, "empty line style pattern",
			 (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Add a trailing background element, if it wasn't specified.
     */
    if (lPtr->pattern[lPtr->plength-1] > 0. ) {
	nbl = length - (ptk - sstring) - strlen(ptk) - 1;
	nbl = nbl > 0 ? nbl : 0;
	lPtr->pattern[lPtr->plength++] = - ((bg[first]+bg[last])/2. +
					nbl * (bi[first]+bi[last])/2.);
    }

#ifdef DEBUG
    {
	int i;
	for ( i=0; i<lPtr->plength; ++i ) {
	    pltMessage("inf", " %.3f", lPtr->pattern[i] );
	}
	pltMessage("inf", "\n");
    }
#endif

    return TCL_OK;
}

int
styFillFG(Tcl_Interp *interp,	/* TCL interpreter, used for error messages */
	  StyleP sty,		/* Style to be modified			    */
	  char *col)		/* Color				    */
/*
 * Set the foreground color of a style.  The color is given in one of
 * several forms:
 *
 * - by name
 *	Currently XD's own database is used to convert color names to
 *      RGB values when necessary.  (I.e. not for TK)
 *
 * - by RGB triple
 *	X Window System style
 *	#rgb, #rrggbb #rrrgggbbb #rrrrggggbbbb   with 'r' being a hex digit
 *
 * - by model+values lists
 *	{RGB red green blue}
 *		red, greeen, blue: [0., 1.]
 *      {HSV hue saturation value}
 *		hue: [0., 360.) 0. is red
 *		saturation: [0., 1.]
 *		value: [0., 1.]
 *	{YUV luminance u v}
 *		luminance, u, v: [0., 1.]
 *		u, v: [-0.5, 0.5]
 */
{
    int i, len, iv[3], iprec;
    int code, elemc;
    char **elemv;

    if (!col) {
	Tcl_AppendResult(interp, "empty color name", (char *) NULL);
	return TCL_ERROR;
    }

    if (col[0] == '#') {
	/*
	 * RGB triple, the X Window System way.
	 * iprec is set to the maximum value that can be specified, using
	 * the given number of bits.
	 */
	len = strlen(col);
	switch (len) {
	  case 4: {
	      sscanf(col, "#%1x%1x%1x", &iv[0], &iv[1], &iv[2]);
	      iprec = 15;
	  }; break;
	  case 7: {
	      sscanf(col, "#%2x%2x%2x", &iv[0], &iv[1], &iv[2]);
	      iprec = 255;
	  }; break;
	  case 10: {
	      sscanf(col, "#%3x%3x%3x", &iv[0], &iv[1], &iv[2]);
	      iprec = 4095;
	  }; break;
	  case 13: {
	      sscanf(col, "#%4x%4x%4x", &iv[0], &iv[1], &iv[2]);
	      iprec = 65535;
	  }; break;
	  default: {
	      Tcl_AppendResult(interp, "illegal color syntax", (char *) NULL);
	      return TCL_ERROR;
	  }
	}
	return styFillColorByRGB(interp, &sty->fg, iprec, iv);
    } else {
	/*
	 * Check for a list of values
	 * NOTE: take care to free elemv later on.
	 */
	if (Tcl_SplitList(interp, col, &elemc, &elemv) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (elemc == 1) {
	    /*
	     * just one value, so it should be a name
	     */
	    ckfree((char *) elemv);
	    return styFillColorByName(interp, &sty->fg, col);
	} else if (elemc != 4) {
	    ckfree((char *) elemv);
	    Tcl_AppendResult(interp, "illegal color syntax", (char *) NULL);
	    return TCL_ERROR;
	}
	/*
	 * four elements, should be model and values
	 */
	for (i=0; i<NCOLMODELS; ++i) {
	    if (strcasecmp(colmodel[i],elemv[0]) == 0) {
		code = styFillColorByValue(interp, &sty->fg,
                                           (ColorE) i, &elemv[1]);
		ckfree((char *) elemv);
		return code;
	    }
	}
	ckfree((char *) elemv);
	Tcl_AppendResult(interp, "illegal color syntax", (char *) NULL);
	return TCL_ERROR;
    }
}

static int
styFillColorByName(Tcl_Interp *interp,	/* TCL interpreter, used for error
					 * messages */
	       ColorP col,		/* color to be changed		*/
	       char *name)		/* color name			*/
{
    /*
     * copy color name
     */
    col->name = strDuplOver(col->name, name);
    /*
     * convert color name to RGB triple
     */
    styCvtColorname(col);
    return TCL_OK;
}

static int
styFillColorByRGB(Tcl_Interp *interp,	/* TCL interpreter, used for error
					 * messages */
		 ColorP col,		/* color to be changed		*/
		 int iprec,		/* maximum RGB value		*/
		 int rgb[3])		/* RGB triple			*/
{
    int i;

    /*
     * set color name and model
     */
    if (col->name) ckfree((char *)col->name);
    col->name = NULL;
    col->model = RGBMODEL;

    /*
     * set and scale values
     */
    for (i=0; i<3; ++i) {
	col->val[i] = (float) rgb[i]/ (float) iprec;
    }
    
    return TCL_OK;
}

static int
styFillColorByValue(Tcl_Interp *interp,	/* TCL interpreter, used for error
					 * messages */
		ColorP col,		/* color to be changed		*/
		ColorE model,		/* color model			*/
		char *val[3])		/* values			*/
{
    int i;
    double value[3];

    /*
     * set color name and model
     */
    if (col->name) ckfree((char *)col->name);
    col->name  = NULL;
    col->model = model;
    /*
     * set values
     */
    for (i=0; i<3; ++i) {
	if ( Tcl_GetDouble(interp, val[i], &value[i]) != TCL_OK ) {
	    return TCL_ERROR;
	}
	col->val[i] = value[i];
    }
    
    return TCL_OK;
}

int
styFillInterpol(Tcl_Interp *interp,
                StyleP sty,
                int step,
                int max,
                StyleP from,
                StyleP to)
{
    float a, b;

    if (from->fg.model != to->fg.model) {
	Tcl_AppendResult(interp, "color models do not match", (char *) NULL);
	return TCL_ERROR;
    }
    memcpy(sty, from, sizeof(StyleT));
    if (max > 1) {
	b = (float) step / (float) (max-1);
	if ( b > 1. ) {
	    b = 1.;
	} else if ( b < 0.) {
	    b = 0.;
	}
	a = 1. - b;
	sty->fg.val[0] = a*from->fg.val[0] + b*to->fg.val[0];
	sty->fg.val[1] = a*from->fg.val[1] + b*to->fg.val[1];
	sty->fg.val[2] = a*from->fg.val[2] + b*to->fg.val[2];
    }
    sty->fg.name=NULL;
    return TCL_OK;
}
