#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/vieManager.c,v 2.11 1996/07/28 11:42:07 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: vieManager.c,v $
 * Revision 2.11  1996/07/28 11:42:07  richter
 * clear display list variables
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/03  10:56:17  richter
 * viewGetConfig and many other changes
 *
 * Revision 2.9.1.1  1996/05/20  11:21:21  richter
 * lots of changes especially viewParseArgs
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:06  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"               /* type definitions and array sizes     */
#include "vieProto.h"
#include "namProto.h"
#include "styProto.h"
#include "strProto.h"
#include "objProto.h"
#include "linProto.h"

ViewP
viewAlloc(char *name)
{
    ViewP view;

    view = (ViewP) ckalloc( sizeof(ViewT) );

    if (! view) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }

    viewClean(view, FALSE);
    view->name = strDupl(name);
    if (namAlloc("VIE", name, (void *)view) != TCL_OK) {
        viewFree(view);
        view = NULL;
    }

    return view;
}

void viewClean(ViewP view,
	      int old)
{
    int i, k;
    
    if (old) {
	if (view->name) ckfree((char *)view->name);
        /*
          if (view->head) ...
          */
    }

    view->name = NIL;
    view->head = NULL;
    view->tail = NULL;
    view->dl_head = NULL;
    view->dl_current = NULL;
    view->dl_tail = NULL;
    view->type = VIEW_UNKNOWN;
    view->opt.doscale = 1;
    view->opt.doorig  = 0;
    view->opt.dotrans = 0;
    view->opt.dorot   = 0;
    view->opt.scale = 1.;
    for (i=0; i<3; ++i) {
        view->opt.orig[i] = 0.;
        for (k=0; k<3; ++k) {
            view->opt.rotm[i][k] = 0.;
        }
        view->opt.rotm[i][i] = 1.;
    }
    view->opt.width = 0.;
    view->opt.height = 0.;
    view->drdat = (void *) 0;
    styClean(&view->style, old);
}

void
viewFree(ViewP view)
{
    if (!view) return;

    namFree("VIE", view->name);
    if (view->name) ckfree((char *)view->name);
    /*
      if (view->head)
      */
    styClean(&view->style, TRUE);
    ckfree((char *)view);
    return;
}

int
viewListByName(char *match)
{
    return namListByName("VIE", match);
}

ViewP
viewFindByName(char *name, char *type)
{
    char *p;
    
    p = viewMakeName(name, type);
    return (ViewP) namFindByName("VIE", p);
}

char *
viewMakeName(char *name, char *type)
{
    static char *p = NULL;

    if ( ! type ) {
	return name;
    }

    if ( p ) {
	ckfree(p);
	p = NULL;
    }
    p = ckalloc(strlen(name)+strlen(type)+1);
    if ( ! p) {
	return NULL;
    }
    strcpy(p, type);
    /*    p[strlen(type)] = '\000'; */
    return strcat(p, name);
}

void
viewRegisterObject(ViewP view,
                   char *name,
                   char *type)
{
    ObjlistP o;

    o = (ObjlistP) ckalloc(sizeof(ObjlistT));
    if (o) {
        if (view->head == NULL || view->tail == NULL) {
            view->head = o;
            o->prev = NULL;
        } else {
            view->tail->next = o;
            o->prev = view->tail;
        }
        o->name = strDupl(name);
        o->type = strDupl(type);
        o->next = NULL;
        view->tail = o;
    }
}

void
viewUnregisterObject(ViewP view,
                     char *name)
{
    ObjlistP p, o;

    o = NULL;
    for (p=view->head; p; p=p->next) {
        if (strcmp(p->name, name)==0) {
            o = p;
            break;
        }
    }

    if (o) {
        if (o->prev) {
            o->prev->next = o->next;
        } else {
            view->head = o->next;
        }
        if (o->next) {
            o->next->prev = o->prev;
        } else {
            view->tail = o->prev;
        }
        ckfree(o->name);
        ckfree(o->type);
        ckfree((char *)o);
    }
}

int
viewParseArgs(ObjectP obj,
              int *argc,
              char **argv,
              ViewOptT *opt)
{
    int i, k, l, skip;
    double value, shift[3], rotm[3][3];
    double vx[3], vy[3], defpnt[3][3];
    int elemc;
    char **elemv = NULL;

    for (skip=i=0; i < *argc; ++i) {
        if (strEQ(argv[i],"-scale", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "scale factor missing", NIL);
		return TCL_ERROR;
	    }
	    if (strcmp(argv[i],"auto")==0) {
		opt->scale = -1.;
	    } else {
		if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		    return TCL_ERROR;
		}
		opt->scale = value;
	    }
            opt->doscale = 1;
        } else if (strEQ(NULL,"-width", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "width missing", NIL);
		return TCL_ERROR;
	    }
            if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
                return TCL_ERROR;
            }
            opt->width = value;
            opt->doscale = 1;
        } else if (strEQ(NULL,"-height", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "height missing", NIL);
		return TCL_ERROR;
	    }
            if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
                return TCL_ERROR;
            }
            opt->height = value;
            opt->doscale = 1;
        } else if (strEQ(NULL,"-vrange", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "range missing", NIL);
		return TCL_ERROR;
            }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
                return TCL_ERROR;
            }
            if (elemc != 2) {
                Tcl_AppendResult(interp, "wrong # of list elements: must be\"",
                                 argv[i-1], " {x-range y-range}\"", NULL);
                goto Error;
            }
            for (k=0; k<2; ++k) {
                if (Tcl_GetDouble(interp, elemv[k], &shift[k]) != TCL_OK) {
                    goto Error;
                }
            }
            for (k=0; k<2; ++k) {
                opt->vrange[k] = shift[k];
            }
            opt->doscale = 1;
        } else if (strEQ(NULL,"-origin", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "origin missing", NIL);
		return TCL_ERROR;
            }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
                return TCL_ERROR;
            }
            if (elemc != 3) {
                Tcl_AppendResult(interp, "wrong # of list elements: must be\"",
                                 argv[i-1], " {orig-x orig-y orig-z}\"", NULL);
                goto Error;
            }
            for (k=0; k<3; ++k) {
                if (Tcl_GetDouble(interp, elemv[k], &shift[k]) != TCL_OK) {
                    goto Error;
                }
            }
            for (k=0; k<3; ++k) {
                opt->orig[k] = shift[k];
            }
            opt->doorig = 1;
        } else if (strEQ(NULL,"-translate", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "translation missing", NIL);
		return TCL_ERROR;
            }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
                return TCL_ERROR;
            }
            for (k=0; k<3; ++k) {
                if (Tcl_GetDouble(interp, elemv[k], &shift[k]) != TCL_OK) {
                    goto Error;
                }
            }
            for (k=0; k<3; ++k) {
                opt->trans[k] = shift[k];
            }
            opt->dotrans = 1;
        } else if (strEQ(NULL,"-matrix", 4)) {
            if (++i >= *argc) {
		Tcl_AppendResult(interp, "rotation matrix missing", NIL);
		return TCL_ERROR;
            }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
                return TCL_ERROR;
            }
            for (k=0; k<3; ++k) {
                for (l=0; l<3; ++l) {
                    if (Tcl_GetDouble(interp, elemv[k*3+l], &rotm[k][l])
                        != TCL_OK) {
                        return TCL_ERROR;
                    }
                }
            }
            for (k=0; k<3; ++k) {
                for (l=0; l<3; ++l) {
                    opt->rotm[k][l] = rotm[k][l];
                }
            }
            opt->dorot = 1;
        } else if (obj && strEQ(NULL,"-3point", 4)) {
            if (++i >= *argc) {
                Tcl_AppendResult(interp, "wrong # args: must be \"",
                                 argv[i-1], " {point-1 point-2 point-3}\"",
                                 NULL);
                return TCL_ERROR;
            }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
                return TCL_ERROR;
            }
            if (elemc != 3) {
                Tcl_AppendResult(interp, "wrong # list elements: must be \"",
                                 argv[i-1], " {point-1 point-2 point-3}\"",
                                 NULL);
                return TCL_ERROR;
            }
            for (k=0; k<3; ++k) {
                if (objGetCoord(interp, elemv[k], obj, defpnt[k])
                    != TCL_OK) {
                    goto Error;
                }
            }
            for (k=0; k<3; ++k) {
                opt->orig[k] = (defpnt[0][k] + defpnt[1][k] + defpnt[2][k])/3.;
                vx[k] = defpnt[1][k] - defpnt[0][k];
                vy[k] = defpnt[2][k] - defpnt[0][k];
            }
            linCoord(opt->rotm, vx, vy);
            opt->doorig = 1;
            opt->dorot  = 1;
        } else {
            argv[skip++] = argv[i];
        }
    }
    *argc = skip;
    if (elemv) {
        ckfree((char *) elemv);
        elemv = NULL;
    }
    return TCL_OK;

Error:
    ckfree((char *) elemv);
    elemv = NULL;
    return TCL_ERROR;
}

int
viewGetConfig(Tcl_DString *o,
              ViewOptT *opt,
              int what)
{
    int i,k;
    
    Tcl_DStringStartSublist(o);
    Tcl_DStringAppendElement(o, "_SECTION");
    Tcl_DStringAppendElement(o, "View Options");
    Tcl_DStringEndSublist(o);

    Tcl_DStringStartSublist(o);
    Tcl_DStringAppendElement(o, "width");
    Tcl_DStringAppendElement(o, "float::create");
    Tcl_DStringAppendElement(o, "Width");
    Tcl_DStringAppendElement(o, "");
    Tcl_DStringStartSublist(o);
    if (opt) {
        sprintf(buf, "%g", opt->width);
        Tcl_DStringAppendElement(o, buf);
    }
    Tcl_DStringEndSublist(o);
    Tcl_DStringEndSublist(o);

    Tcl_DStringStartSublist(o);
    Tcl_DStringAppendElement(o, "height");
    Tcl_DStringAppendElement(o, "float::create");
    Tcl_DStringAppendElement(o, "Height");
    Tcl_DStringAppendElement(o, "");
    Tcl_DStringStartSublist(o);
    if (opt) {
        sprintf(buf, "%g", opt->height);
        Tcl_DStringAppendElement(o, buf);
    }
    Tcl_DStringEndSublist(o);
    Tcl_DStringEndSublist(o);

    Tcl_DStringStartSublist(o);
    Tcl_DStringAppendElement(o, "scale");
    Tcl_DStringAppendElement(o, "float");
    Tcl_DStringAppendElement(o, "Scale Factor");
    Tcl_DStringAppendElement(o, "");
    if (opt && opt->doscale) {
        sprintf(buf, "%g", opt->scale);
        Tcl_DStringAppendElement(o, buf);
    } else {
        Tcl_DStringAppendElement(o, "");
    }
    Tcl_DStringEndSublist(o);

    if (what&2) {
        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "vrange");
        Tcl_DStringAppendElement(o, "float:2");
        Tcl_DStringAppendElement(o, "Data Value Range");
        Tcl_DStringAppendElement(o, "");
        Tcl_DStringStartSublist(o);
        if (opt) {
            for (i=0; i<2; ++i) {
                sprintf(buf, "%g", opt->vrange[i]);
                Tcl_DStringAppendElement(o, buf);
            }
        }
        Tcl_DStringEndSublist(o);
        Tcl_DStringEndSublist(o);
    }
        
    if (what&1) {
        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "origin");
        Tcl_DStringAppendElement(o, "float:3");
        Tcl_DStringAppendElement(o, "Origin");
        Tcl_DStringAppendElement(o, "");
        Tcl_DStringStartSublist(o);
        if (opt && opt->doorig) {
            for (i=0; i<3; ++i) {
                sprintf(buf, "%g", opt->orig[i]);
                Tcl_DStringAppendElement(o, buf);
            }
        }
        Tcl_DStringEndSublist(o);
        Tcl_DStringEndSublist(o);

        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "matrix");
        Tcl_DStringAppendElement(o, "float:9");
        Tcl_DStringAppendElement(o, "Rotation");
        Tcl_DStringAppendElement(o, "");
        Tcl_DStringStartSublist(o);
        if (opt && opt->dorot) {
            for (i=0; i<3; ++i) {
                for (k=0; k<3; ++k) {
                    sprintf(buf, "%g", opt->rotm[i][k]);
                    Tcl_DStringAppendElement(o, buf);
                }
            }
        }
        Tcl_DStringEndSublist(o);
        Tcl_DStringEndSublist(o);

        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "translate");
        Tcl_DStringAppendElement(o, "float:3");
        Tcl_DStringAppendElement(o, "Translation");
        Tcl_DStringAppendElement(o, "");
        Tcl_DStringStartSublist(o);
        if (opt && opt->dotrans) {
            for (i=0; i<3; ++i) {
                sprintf(buf, "%g", opt->trans[i]);
                Tcl_DStringAppendElement(o, buf);
            }
        }
        Tcl_DStringEndSublist(o);
        Tcl_DStringEndSublist(o);
    }

    return TCL_OK;

}
