#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/CSBox_3.tcl,v 1.6 1998/11/18 12:36:00 richter Rel $
#
#
# $Log: CSBox_3.tcl,v $
# Revision 1.6  1998/11/18 12:36:00  richter
# incr only works on integers, replace with expr x+1
#
# Revision 1.5  1996/10/11  11:05:03  richter
# the 1.4 fixes did not work.
#
# Revision 1.4  1996/09/23  12:04:58  richter
# cancel returned the mode string -- fixed
# -preset <colorname> did not work -- partially fixed
#
# Revision 1.3  1996/09/21  11:49:07  richter
# corrected the label for 'rgb' mode
#
# Revision 1.2  1996/09/06  12:26:57  richter
# initialize mode
#
# Revision 1.1  1996/05/23  07:42:08  richter
# Initial revision
#
#

proc CSBox {root args} {
    global csBox

    set mode rgb
    set red   0.3
    set green 0.3
    set blue  0.3
    set w [toplevel $root.colorselect]
    set widgets {}
    set modal 1
    while {[llength $args] >= 2} {
	set option [lindex $args 0]
	set value [lindex $args 1]
	switch -- $option {
	    -widgets {
		set widgets $value
	    }
	    -preset {
		set mode [string tolower [lindex $value 0]]
		set value [lreplace $value 0 0]
		switch -- $mode {
		    hsv {
			set value [eval CSBox.hsv2rgb $value]
		    }
		    yuv {
			set value [eval CSBox.yuv2rgb $value]
		    }
		    rgb {
		    }
		    default {
			set mode "rgb"
			set value {0.3 0.3 0.3}
		    }
		}
		set red [lindex $value 0]
		set green [lindex $value 1]
		set blue [lindex $value 2]
	    }
	    -modal {
		set modal $value
	    }
	}
	set args [lreplace $args 0 1]
    }
    
#
# modi
#
    set f [frame $w.modi]
    radiobutton $f.rgb -text rgb -value rgb -variable csBox($w.slider.f) \
	-command [list CSBox.rgbslider $w.slider $w.colorview [concat $w.colorview $widgets]]
    radiobutton $f.hsv -text hsv -value hsv -variable csBox($w.slider.f) \
	-command [list CSBox.hsvslider $w.slider $w.colorview [concat $w.colorview $widgets]]
    radiobutton $f.yuv -text yuv -value yuv -variable csBox($w.slider.f) \
	-command [list CSBox.yuvslider $w.slider $w.colorview [concat $w.colorview $widgets]]
    pack $f.rgb $f.hsv $f.yuv -side left -expand yes
#
# generic frame for rgb, hsv or yuv slider
#
    frame $w.slider 
#
# color display
#
    frame $w.colorview -width 100 -height 50 -background [format "#%02x%02x%02x" [expr int($red*255.)] [expr int($green*255.)] [expr int($blue*255.)]] -bd 2 -relief raised
#
# external buttons
#
    set f [frame $w.buttons]
    button $f.ok -text Ok -width 6 -command "CSBox.ok $w"
    button $f.cancel -text Cancel -width 6 -command "CSBox.cancel $w"
    pack $f.ok $f.cancel  -side left -expand yes
#
    $w.modi.$mode invoke
    pack $w.modi $w.slider $w.colorview $w.buttons \
	-side top -expand yes -fill both -pady 5
#
    if {$modal} {
	tkwait window $w
	return $csBox($w)
    }
}

proc CSBox.ok {w} {
    global csBox

    set csBox($w) "[string toupper $csBox($w.slider.f)] [CSBox.get $w.slider.f $csBox($w.slider.f)]"
    destroy $w
}

proc CSBox.cancel {w} {
    global csBox

    set csBox($w) ""
    destroy $w
}

proc CSBox.rgbslider {frame source widgets} {
    set w $frame.f
    if {[winfo exists $w]} {
	destroy $w
    }
    frame $w
    scale $w.red   -command "CSBox.set $w \"$widgets\"" \
	-label red   -showvalue yes\
	-from 0 -to 255 -length 360 -orient horizontal
    scale $w.green -command "CSBox.set $w \"$widgets\"" \
	-label green -showvalue yes\
	-from 0 -to 255 -length 360 -orient horizontal
    scale $w.blue  -command "CSBox.set $w \"$widgets\"" \
	-label blue  -showvalue yes\
	-from 0 -to 255 -length 360 -orient horizontal

    set rgb [CSBox.getbg $source]
    $w.red set [expr int([lindex $rgb 0]*255.)]
    $w.green set [expr int([lindex $rgb 1]*255.)]
    $w.blue set [expr int([lindex $rgb 2]*255.)]

    pack $w.red $w.green $w.blue -side top
    pack $w
}

proc CSBox.hsvslider {frame source widgets} {
    set w $frame.f
    if {[winfo exists $w]} {
	destroy $w
    }
    frame $w
    scale $w.hue   -command "CSBox.set $w \"$widgets\"" \
	-label hue   -showvalue yes \
	-from 0 -to 360 -length 360 -orient horizontal
    scale $w.sat -command "CSBox.set $w \"$widgets\"" \
	-label saturation -showvalue yes \
	-from 0 -to 100 -length 360 -orient horizontal
    scale $w.val  -command "CSBox.set $w \"$widgets\"" \
	-label value  -showvalue yes \
	-from 0 -to 100 -length 360 -orient horizontal

    set hsv [eval CSBox.rgb2hsv [CSBox.getbg $source]]
    $w.hue set [expr int([lindex $hsv 0]*360.)]
    $w.sat set [expr int([lindex $hsv 1]*100.)]
    $w.val set [expr int([lindex $hsv 2]*100.)]
    pack $w.hue $w.sat $w.val -side top
    pack $w

}
proc CSBox.yuvslider {frame source widgets} {
    set w $frame.f
    if {[winfo exists $w]} {
	destroy $w
    }
    frame $w
    scale $w.y -command "CSBox.set $w \"$widgets\"" \
	-label y -showvalue yes \
	-from 0 -to 100 -length 360 -orient horizontal
    scale $w.u -command "CSBox.set $w \"$widgets\"" \
	-label u -showvalue yes \
	-from 0 -to 100 -length 360 -orient horizontal
    scale $w.v -command "CSBox.set $w \"$widgets\"" \
	-label v -showvalue yes \
	-from 0 -to 100 -length 360 -orient horizontal

    set yuv [eval CSBox.rgb2yuv [CSBox.getbg $source]]
    $w.y set [expr int([lindex $yuv 0]*100.)]
    $w.u set [expr int([lindex $yuv 1]*100.)]
    $w.v set [expr int([lindex $yuv 2]*100.)]
    pack $w.y $w.u $w.v -side top
    pack $w

}

proc CSBox.get {w {as ""}} {
    global csBox

    switch -- $csBox($w) {
	hsv {
	    set rgb [CSBox.hsv2rgb [expr [$w.hue get]/360.] [expr [$w.sat get]/100.] [expr [$w.val get]/100.]]
	}
	yuv {
	    set rgb [CSBox.yuv2rgb [expr [$w.y get]/100.] [expr [$w.u get]/100.] [expr [$w.v get]/100.]]
	}
	rgb -
	default {
	    set rgb "[expr [$w.red get]/255.] [expr [$w.green get]/255.] [expr [$w.blue get]/255.]"
	}
    }

    switch -- $as {
	hsv {
	    return [eval CSBox.rgb2hsv $rgb]
	}
	yuv {
	    return [eval CSBox.rgb2yuv $rgb]
	}
	rgb {
	    return $rgb
	}
	default {
	    return [format "#%02x%02x%02x" [expr int([lindex $rgb 0]*255.)] [expr int([lindex $rgb 1]*255.)] [expr int([lindex $rgb 2]*255.)]]
	}
    }
}

proc CSBox.set {root widgets intensity} {
    set rgb [CSBox.get $root]
    foreach w $widgets {
	$w configure -background $rgb
    }
}

proc CSBox.getbg {w} {
    set color [lindex [$w configure -background] 4]
    scan $color "#%02x%02x%02x" r g b
    return "[expr $r/255.] [expr $g/255.] [expr $b/255.]"
}

proc CSBox.max3 {a b c} {
    if {$a > $b} {
	if {$a > $c} {
	    return $a
	} else {
	    return $c
	}
    } else {
	if {$b > $c} {
	    return $b
	} else {
	    return $c
	}
    }
}

proc CSBox.min3 {a b c} {
    if {$a < $b} {
	if {$a < $c} {
	    return $a
	} else {
	    return $c
	}
    } else {
	if {$b < $c} {
	    return $b
	} else {
	    return $c
	}
    }
}

proc CSBox.hsv2rgb {h s v} {

    if {$s == 0.} {
	set r $v
	set g $v
	set b $v
    } else {
	while {$h < 0.} {set h [expr $h+1.]}
	while {$h >= 1.} {set h [expr $h-1.]}
	set h [expr $h*6.]
	set f [expr $h - int($h)]
	set p [expr $v * (1 - $s)]
	set q [expr $v * (1 - ($s*$f))]
	set t [expr $v * (1 - ($s*(1-$f)))]
	switch [expr int($h)] {
	    0 { set r $v; set g $t; set b $p}
	    1 { set r $q; set g $v; set b $p}
	    2 { set r $p; set g $v; set b $t}
	    3 { set r $p; set g $q; set b $v}
	    4 { set r $t; set g $p; set b $v}
	    5 { set r $v; set g $p; set b $q}
	}
    }
    return "$r $g $b"
}

proc CSBox.rgb2hsv {r g b} {
    set max [CSBox.max3 $r $g $b]
    set min [CSBox.min3 $r $g $b]
    set delta [expr $max - $min]
    set v $max
    if {$max != 0  && $delta != 0} {
	set s [expr $delta/$max]
	if {$r == $max} {
	    set h [expr ($g - $b) / $delta]
	} elseif {$g == $max} {
	    set h [expr 2 + ($b - $r) / $delta]
	} else {
	    set h [expr 4 + ($r - $g) / $delta]
	}
	set h [expr $h/6.]
        while {$h < 0.} {
            set h [expr $h+1]
	}
    } else {
	set s 0.
	set h 0.
    }
    return "$h $s $v"
}

proc CSBox.trimyuv {y u v a b} {
    set au [expr $a * $u]
    set ybv [expr -$y - $b * $v]
    
    if {$au < $ybv} {
        if {$u == 0.} {
            set v [expr -$y / $b]
        } else {
            set vu [expr $v / $u]
            set u [expr -$y / ($a + $b * $vu)]
            set v [expr $vu * $u]
        }
    } else {
	if {$au > 1. + $ybv} {
	    if {$u == 0.} {
		set v [expr (1. - $y) / $b]
	    } else {
		set vu [expr $v / $u]
		set u [expr (1. - $y) / ($a + $b * $vu)]
		set v [expr $vu * $u]
	    }
	}
    }
    return "$u $v"
}

proc CSBox.yuv2rgb {y uu vv} {
    set u [expr $uu-0.5]
    set v [expr $vv-0.5]

    set m {{ 1.00000000  0.00000000  1.40200007} \
	       { 1.00000000 -0.34413627 -0.71413624} \
	       { 1.00000000  1.77199996  0.00000000}}
    set uv [CSBox.trimyuv $y $u $v [lindex [lindex $m 0] 1] \
		[lindex [lindex $m 0] 2]]
    set u [lindex $uv 0]
    set v [lindex $uv 1]
    set uv [CSBox.trimyuv $y $u $v [lindex [lindex $m 1] 1] \
		[lindex [lindex $m 1] 2]]
    set u [lindex $uv 0]
    set v [lindex $uv 1]
    set uv [CSBox.trimyuv $y $u $v [lindex [lindex $m 2] 1] \
		[lindex [lindex $m 2] 2]]
    set u [lindex $uv 0]
    set v [lindex $uv 1]

    set r [expr $y + $u*[lindex [lindex $m 0] 1] + $v*[lindex [lindex $m 0] 2]]
    set g [expr $y + $u*[lindex [lindex $m 1] 1] + $v*[lindex [lindex $m 1] 2]]
    set b [expr $y + $u*[lindex [lindex $m 2] 1] + $v*[lindex [lindex $m 2] 2]]
    return "$r $g $b"
}

proc CSBox.rgb2yuv {r g b} {

    set m {{ 0.29900000 0.58700000 0.11400000} \
	       {-0.16873589 -0.33126411  0.50000000} \
	       { 0.50000000 -0.41868759 -0.08131241}}

    set y [expr $r*[lindex [lindex $m 0] 0] + \
	       $g*[lindex [lindex $m 0] 1]+ \
	       $b*[lindex [lindex $m 0] 2]]
    set u [expr $r*[lindex [lindex $m 1] 0] + \
	       $g*[lindex [lindex $m 1] 1] + \
	       $b*[lindex [lindex $m 1] 2]]
    set v [expr $r*[lindex [lindex $m 2] 0] + \
	       $g*[lindex [lindex $m 2] 1] + \
	       $b*[lindex [lindex $m 2] 2]]
    return "$y [expr $u+0.5] [expr $v+0.5]"
};

