#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/FSBox_3.tcl,v 1.4 1996/11/05 15:14:31 richter Rel $
#
#
# $Log: FSBox_3.tcl,v $
# Revision 1.4  1996/11/05 15:14:31  richter
# (setselection): don't use the dirname bug which got fixed in Tcl 7.5
#
# Revision 1.3  1996/09/22  14:52:29  richter
# clear _fsbox(argfilter) when called
#
# Revision 1.2  1996/09/06  12:35:05  richter
# use proc exists instead of lsearch [array names
# bindings for <Return>
# change cursor while scaning the directory
#
# Revision 1.1  1996/05/23  07:42:08  richter
# Initial revision
#
#

proc FSBox {root args} {
    global _fsbox _fsbox_cache

    set modal 1
    set _fsbox(pwd) "pwd"
    set _fsbox(argfilter) ""
    while {[llength $args] >= 2} {
        set option [lindex $args 0]
        set value [lindex $args 1]
        switch -- $option {
            -modal {
                set modal $value
            }
	    -pwd {
		set _fsbox(pwd) $value
	    }
	    -filter {
		set _fsbox(argfilter) $value
	    }
	    default {
		puts stderr "unknown option: $option"
	    }
        }
        set args [lreplace $args 0 1]
    }
    #
    # initialize _fsbox
    #
    if {![exists _fsbox path] || $_fsbox(path) == ""} {
	set cmd [list exec sh -c "$_fsbox(pwd)"]
	if {[catch $cmd path]} {
	    puts stderr "$cmd: $path"
	    return
	}
	set _fsbox(path) $path
    }
    if {![exists _fsbox argfilter] || $_fsbox(argfilter) == ""} {
	set _fsbox(filter) "$_fsbox(path)/*"
    } else {
	set _fsbox(filter) "$_fsbox(path)/$_fsbox(argfilter)"
    }
    set _fsbox(selection) "$_fsbox(path)"
    #
    # make sure _fsbox_cache is an array
    #
    set _fsbox_cache() ""

    set w [toplevel $root.fileselect -borderwidth 5]
    set _fsbox(root) $w
    #
    # raw layout
    #
    frame $w.filter -borderwidth 5
    frame $w.ls -borderwidth 5
    frame $w.selection -borderwidth 5
    frame $w.buttons -borderwidth 5
    #
    # filter
    #
    label $w.filter.label -text Filter -anchor sw
    entry $w.filter.entry -textvariable _fsbox(filter)
    pack $w.filter.label $w.filter.entry -side top -expand yes -fill both
    bind $w.filter.entry <Return> _fsbox.filter
    #
    # directories
    #
    set dir [frame $w.ls.dir]
    _fsbox.ssl $dir.ssl
    label $dir.label -text Directories -anchor sw
    pack $dir.label $dir.ssl -side top -expand yes -fill both
    #
    # file
    #
    set file [frame $w.ls.file]
    label $file.label -text Files -anchor sw
    _fsbox.ssl $file.ssl
    pack $file.label $file.ssl -side top -expand yes -fill both
    #
    # ls
    #
    pack $dir [frame $w.ls.spacer -width 10] $file -side left -expand yes
    set _fsbox(file_lb) $file.ssl.vsl.list
    set _fsbox(dir_lb) $dir.ssl.vsl.list
    bind $dir.ssl.vsl.list <Double-1> _fsbox.filter
    bind $dir.ssl.vsl.list <ButtonRelease-1> _fsbox.setfilter
    bind $file.ssl.vsl.list <ButtonRelease-1> _fsbox.setselection
    #
    # selection
    #
    label $w.selection.label -text Selection -anchor sw
    entry $w.selection.entry -textvariable _fsbox(selection)
    pack $w.selection.label $w.selection.entry -side top -expand yes -fill both
    bind $w.selection.entry <Return> "_fsbox.ok $w"
    #
    # buttons
    #
    button $w.buttons.ok -text OK -width 6 -padx 2 \
	-command "_fsbox.ok $w"
    button $w.buttons.filter -text Filter -width 6 -padx 2 \
	-command _fsbox.filter
    button $w.buttons.cancel -text Cancel -width 6 -padx 2 \
	-command "_fsbox.cancel $w"
    pack $w.buttons.ok $w.buttons.filter $w.buttons.cancel -side left -expand yes
    #
    # everything
    #
    pack $w.filter $w.ls $w.selection -expand yes -fill x
    #    pack [frame $w.ridge -relief sunken -borderwidth 1 -height 2] -side top -fill x -expand yes
    pack $w.buttons -side top -expand yes -fill x
    update
    _fsbox.filter
    if {$modal} {
	tkwait window $w
	return $_fsbox(selection)
    }
}

proc _fsbox.ssl {w} {
    global tk_version

    frame $w

    frame $w.vsl
    listbox $w.vsl.list -exportselection no \
	-yscrollcommand "$w.vsl.scroll set" \
	-xscrollcommand "$w.hsl.scroll set"
    if {$tk_version < 4.0} {
	 tk_listboxSingleSelect $w.vsl.list
    }
    scrollbar $w.vsl.scroll -relief sunken -width 15 -borderwidth 2 \
	-command "$w.vsl.list yview"
    pack $w.vsl.list -side left
    pack $w.vsl.scroll -side right -expand yes -fill both

    frame $w.hsl
    scrollbar $w.hsl.scroll -relief sunken -width 15 -borderwidth 2 \
	-orient horizontal -command "$w.vsl.list xview"
    frame $w.hsl.filler -width 19
    pack $w.hsl.scroll -side left -expand yes -fill both
    pack $w.hsl.filler -side right

    pack $w.vsl $w.hsl -side top -expand yes -fill both
    return $w
}

proc _fsbox.setfilter {} {
    global _fsbox

    set wd $_fsbox(dir_lb)
    set index [$wd curselection]
    if {"$index" == ""} {
	return
    }
    set dir [$wd get $index]

    set cmd [list exec sh -c "cd $_fsbox(path)/$dir; $_fsbox(pwd)"]
    if {[catch $cmd path]} {
	puts stderr "$cmd: $path"
	return
    }
    set filter [file tail $_fsbox(filter)]
    set _fsbox(filter) "$path/$filter"
}

proc _fsbox.setselection {} {
    global _fsbox

    set wf $_fsbox(file_lb)
    set index [$wf curselection]
    if {"$index" == ""} {
	return
    }
    set file [$wf get $index]
    if {[file isdirectory $_fsbox(selection)]} {
	set path $_fsbox(selection)
    } else {
	set path [file dirname $_fsbox(selection)]
    }
    set _fsbox(selection) "$path/$file"
}

proc _fsbox.filter {} {
    global _fsbox _fsbox_cache

    set path "[file dirname $_fsbox(filter)]"
    if {![file isdirectory $path]} {
	puts stderr "beep-beep"
	return
    }

    set _fsbox(path) $path
    set _fsbox(selection) "$path/"

    $_fsbox(root) configure -cursor watch
    update
    set mtime [file mtime $path]
    if {![exists _fsbox_cache $path] || 
	[lindex $_fsbox_cache($path) 0] < $mtime} {
	if {[catch "exec ls -a $path" ls]} {
	    puts stderr "$ls"
	    $_fsbox(root) configure -cursor ""
	    return
	}
	set _fsbox_cache($path) [list $mtime]
	foreach file [lsort $ls] {
	    lappend _fsbox_cache($path) [list $file [file type $path/$file]]
	}
    }

    set wd $_fsbox(dir_lb)
    set wf $_fsbox(file_lb)
    $wd delete 0 end
    $wf delete 0 end
    set filter [file tail $_fsbox(filter)]
    foreach entry $_fsbox_cache($path) {
	if {[llength $entry] == 2} {
	    set file [lindex $entry 0]
	    set type [lindex $entry 1]
	    if {[string compare $type directory] == 0} {
		$wd insert end $file
	    } else {
		if {[string match $filter $file]} {
		    $wf insert end $file
		}
	    }
	}
    }
    $_fsbox(root) configure -cursor ""
}

proc _fsbox.cancel {w} {
    global _fsbox
    set _fsbox(selection) ""
    destroy $w
}

proc _fsbox.ok {w} {
#    global _fsbox
#
#   if {[string compare $_fsbox(selection) "$_fsbox(path)/"] == 0} {
#	set _fsbox(selection) ""
#    }
    destroy $w
}
