#
# Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.
#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/xdTK.tcl,v 2.4 1999/06/03 12:36:27 richter Exp $
#
#
# $Log: xdTK.tcl,v $
# Revision 2.4  1999/06/03 12:36:27  richter
# first throw at removable views
#
# Revision 2.3  1996/11/11 16:45:45  richter
# ps compatibility proc nedded an eval
#
# Revision 2.2  1996/11/06  13:25:58  richter
# fixed the hardcopy procedure
#
# Revision 2.1  1996/10/11  12:10:30  richter
# added the contents of xdTKmkItem and xdTKview
# new hardcopy routine implementing a unified interface for gt and tk drive
#
# Revision 2.0  1995/05/12  11:54:40  richter
# First semi-public beta release
#
#

scan $tk_version "%d.%d" xd_tk_major xd_tk_minor
if {($xd_tk_major < 3) || (($xd_tk_major == 3) && ($xd_tk_minor < 4))} {
    error "wrong version of Tk loaded ($tk_version): need 3.4 or later"
}

#set xd_library /user/richter/src/xd/xdgraph/tcl

source $xd_library/cli.tcl
source $xd_library/gui.tcl

mkGUI

rename isosurface {}
rename heightfield {}

set active(view) ""
set active(display) ""
set theWidth("") 1
set theHeight("") 1

proc CreateView {view width height} {
    global xd_tk_major

    set w [view2widget $view]
    set d [view2display $view]

    set dwidth  [expr $width + 1]
    set dheight [expr $height + 1]
    set hwidth  [expr $dwidth/2.]
    set hheight [expr $dheight/2.]

    set options "-width ${dwidth}c -height ${dheight}c -scrollregion \
	\"-${hwidth}c -${hheight}c ${hwidth}c ${hheight}c\""

    if {[winfo exists $w]} {
	eval $d configure $options
	ClearView $view
    } else {
	toplevel $w

##	puts $options
	if {$xd_tk_major < 4} {
	    eval canvas $d -relief flat $options -background white \
		-scrollincrement 1m
	    $d xview 0
	    $d yview 0
	} else {
	    eval canvas $d -relief flat $options -background white
	    $d xview moveto 0
	    $d yview moveto 0
	}

	wm title $w "view: $view"
	wm iconname $w "$view"
	pack append $w $d {top}

	mkItems $d

#	bind $w <Control-q> {destroy .}
#	bind $w <Control-c> {destroy .}
#	wm protocol $w WM_DELETE_WINDOW {destroy .}
	bind $w <Destroy>   "remove $view"
	
    }

    DrawDisplayBorder $d $w
    update
    return
}

proc RescaleView {view width height} {

    set w [view2widget $view]
    set d [view2display $view]

    set dwidth  [expr $width + 1]
    set dheight [expr $height + 1]
    set hwidth  [expr $dwidth/2.]
    set hheight [expr $dheight/2.]

    set options "-width ${dwidth}c -height ${dheight}c -scrollregion \
	\"-${hwidth}c -${hheight}c ${hwidth}c ${hheight}c\""

    eval $d configure $options
    ClearView $view
}

proc DrawDisplayBorder {d w} {

    set border [winfo pixels $w 0.5c]
    set region [lindex [$d configure -scrollregion] 4]
    set x0 [expr [winfo pixels $w [lindex $region 0]] + $border - 1]
    set y0 [expr [winfo pixels $w [lindex $region 1]] + $border - 1]
    set x1 [expr [winfo pixels $w [lindex $region 2]] - $border + 1]
    set y1 [expr [winfo pixels $w [lindex $region 3]] - $border + 1]

    $d create rectangle $x0 $y0 $x1 $y1 -width 0.2m
}

proc RemoveView {view} {
    set w [view2widget $view]
    if {[winfo exists $w]} {
	destroy $w
    }
    return
}

proc SetActiveView {view} {
    global active plotcmd

    set active(view) $view
    set active(display) [view2display $view]
    set plotcmd $active(display)
}

proc ClearView {view} {
    global active

    if {$view == ""} {
	set d $active(display)
    } else {
	set d [view2display $view]
    }

    $d delete all
    DrawDisplayBorder $d [view2widget $view]
    return
}

proc ClearObject {view obj} {
    global active

    if {$view == ""} {
	set d $active(display)
    } else {
	set d [view2display $view]
    }
    if {[info commands $d] != {}} {
	$d delete $obj
    }
    return
}

proc mkItems {c} {

# Set up event bindings for object labels on canvas:

    $c bind label <Any-Enter> "itemEnter $c"
    $c bind label <Any-Leave> "itemLeave $c"
    $c bind label <1> "itemStartDrag $c %x %y"
    $c bind label <B1-Motion> "itemDrag $c %x %y"
}

# Utility procedures for highlighting the item under the pointer:

proc itemEnter {c} {
    global restoreCmd

#    if {[tk colormodel $c] != "color"} {
#	set restoreCmd {}
#	return
#    }
    set fill [lindex [$c itemconfig current -fill] 4]
    set restoreCmd "$c itemconfig current -fill $fill"
    $c itemconfig current -fill SteelBlue2
}

proc itemLeave {c} {
    global restoreCmd

    eval $restoreCmd
}


#set areaX1 0
#set areaY1 0
#set areaX2 0
#set areaY2 0

# Utility procedures to support dragging of items.

proc itemStartDrag {c x y} {
    global lastX lastY
    set lastX [$c canvasx $x]
    set lastY [$c canvasy $y]
}

proc itemDrag {c x y} {
    global lastX lastY
    set x [$c canvasx $x]
    set y [$c canvasy $y]
    $c move current [expr $x-$lastX] [expr $y-$lastY]
    set lastX $x
    set lastY $y
}

proc hardcopy {args} {
    global active

    set file ""
    set cmd [list $active(display) postscript -colormode color]
    while {[llength $args] >= 2} {
        set option [lindex $args 0]
        set value  [lindex $args 1]
        set args [lreplace $args 0 1]
        switch -glob -- $option {
            -width {
		lappend cmd -pagewidth $value
	    }
	    -height {
		lappend cmd -pageheight $value
	    }
	    -file {
		set file 1
		lappend cmd -file $value
	    }
	}
    }

    if {$file == ""} {
	lappend cmd -file xdgraph.ps
    }

    if {$active(display) != ""} {
	eval $cmd
    }
}

proc ps {args} {
    eval [concat hardcopy $args]
}
